;==============================================================================

; file: "_t-c.scm"

; Copyright (C) 1994, 1995 by Marc Feeley, All Rights Reserved.

(include "generic.scm")

;------------------------------------------------------------------------------
;
; Back end for C language
; -----------------------

; Back end parameters:

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** REGISTER USAGE

(define targ-nb-gvm-regs 5) ; Number of registers in the virtual machine.
(define targ-nb-arg-regs 3) ; Number of registers used to pass arguments.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** SUPPORTED WORD SIZES (in characters per word)

; It is assumed that the target machine's word size (in characters)
; is a power of 2.  The minimal supported word size must be known to
; decide if an integer should be represented with a fixnum or bignum.  The
; largest supported word size is needed in order to generate portable macros
; to pack strings and bignums into words.

(define targ-min-word-size 4) ; At least 4 characters per word
(define targ-max-word-size 8) ; Up to 8 characters per word

(define targ-bits-per-char 8) ; 8 bits to a character

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** DATA REPRESENTATION

; Note: The data representation is mostly machine independent.  This means
; that it is possible to change the format of objects by redefining the
; appropriate macros in "gambit.h".  However, the back end must know
; a few things about the representation in order for the output to be
; portable to different word sizes and to perform a few optimizations.
; In the following definitions, space is expressed in number of words.

; Lower bound on free space in heap at the beginning of basic blocks:

(define targ-heap-fudge 1024)

; Number of tag bits per pointer:

(define targ-tag-bits 3)
(define targ-alignment (expt 2 targ-tag-bits))

; Upper bound on space needed by various objects:

(define (targ-max-words i) ; minimum k such that targ-min-word-size*n >= i
                           ; and targ-min-word-size*n mod targ-alignment = 0

  (define (round-up-to-multiple-of n i)
    (* (quotient (+ i (- n 1)) n) n))

  (quotient (round-up-to-multiple-of
              targ-min-word-size
              (round-up-to-multiple-of targ-alignment i))
            targ-min-word-size))

(define targ-cons-space        (targ-max-words (* 2 targ-min-word-size)))
(define targ-cell-space        (targ-max-words (* 2 targ-min-word-size)))
(define targ-flonum-space      (targ-max-words 8))
(define targ-placeholder-space (targ-max-words (* 2 targ-min-word-size)))
(define (targ-closure-space n) (targ-max-words (* (+ n 2) targ-min-word-size)))
(define (targ-string-space n)  (targ-max-words (+ (+ n 1) targ-min-word-size)))
(define (targ-vector8-space n) (targ-max-words (+ n targ-min-word-size)))
(define (targ-vector-space n)  (targ-max-words (* (+ n 1) targ-min-word-size)))

; Limits of fixnums (assumes 2's complement representation):

(define targ-max-fixnum
  (- (expt 2
           (- (* targ-min-word-size targ-bits-per-char)
              (+ targ-tag-bits 1)))
     1))

(define targ-min-fixnum
  (- -1 targ-max-fixnum))

; Radix for bignum digits:

; Radix must be a power of two and (targ-bignum-radix - 1)^2 must fit in
; a fixnum.

(define targ-bignum-radix
  (expt 2
        (quotient (- (* targ-min-word-size targ-bits-per-char)
                     (+ targ-tag-bits 1))
                  2)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** PROCEDURE CALLING CONVENTION

(define (targ-label-info min-args nb-parms rest? closed?)

; After a LABEL instruction, the following is true:
;
;  * return address is in R0
;
;  * if nb-parms <= targ-nb-arg-regs,
;
;      then, parameter `n' is in Rn
;
;      else, the first `m' = nb-parms - targ-nb-arg-regs
;            parameters will be on the stack and, assuming `i' = n - m,
;            parameter `n' is in
;
;            Ri, if i > 0
;            or else in slot `frame_size + i' of the stack frame
;
;  * if `CLOSED' is present, register number `targ-nb-arg-regs + 1' contains
;    a pointer to the closure object
;
; for example, if we assume that targ-nb-arg-regs = 3, then after the
; instruction LABEL(1,2,PROC,5):
;
;   R0 = return address
;   S1 = parameter 1
;   S2 = parameter 2
;   R1 = parameter 3
;   R2 = parameter 4
;   R3 = parameter 5

  (let ((nb-stacked (max 0 (- nb-parms targ-nb-arg-regs))))

    (define (location-of-parms i)
      (if (> i nb-parms)
        '()
        (cons (cons i
                    (if (> i nb-stacked)
                      (make-reg (- i nb-stacked))
                      (make-stk i)))
              (location-of-parms (+ i 1)))))

    (let ((x (cons (cons 'return 0) (location-of-parms 1))))
      (make-pcontext nb-stacked
        (if closed?
          (cons (cons 'closure-env (make-reg (+ targ-nb-arg-regs 1))) x)
          x)))))

(define (targ-jump-info nb-args)

; After a JUMP instruction, the following is true:
;
;  * the return address is passed in R0
;
;  * if nb-args <= targ-nb-arg-regs,
;
;      then, argument `n' is in Rn
;
;      else, `m' = nb-args - targ-nb-arg-regs arguments are passed
;            on the stack and, assuming `i' = n - m, argument `n' is in
;
;            Ri, if i > 0
;            or else in slot `frame_size + i' of the stack frame

  (let ((nb-stacked (max 0 (- nb-args targ-nb-arg-regs))))

    (define (location-of-args i)
      (if (> i nb-args)
        '()
        (cons (cons i
                    (if (> i nb-stacked)
                      (make-reg (- i nb-stacked))
                      (make-stk i)))
              (location-of-args (+ i 1)))))

    (make-pcontext nb-stacked
                   (cons (cons 'return (make-reg 0))
                         (location-of-args 1)))))

;------------------------------------------------------------------------------
;
; Initialization/finalization

(let ((targ (make-target 4 'C)))

  (define (begin! info-port)

    (set! targ-info-port info-port)
    (set! targ-return-reg (make-reg 0))

    (target-end!-set!         targ end!)
    (target-dump-set!         targ targ-dump)
    (target-nb-regs-set!      targ targ-nb-gvm-regs)
    (target-prim-info-set!    targ targ-prim-info)
    (target-label-info-set!   targ targ-label-info)
    (target-jump-info-set!    targ targ-jump-info)
    (target-proc-result-set!  targ (make-reg 1))
    (target-task-return-set!  targ targ-return-reg)

    #f)

  (define (end!)
    #f)

  (target-begin!-set! targ begin!)
  (put-target targ))

(define targ-info-port #f)
(define targ-return-reg #f)

;------------------------------------------------------------------------------
;
; Primitive procedure database

(define targ-prim-proc-table
  (map (lambda (x)
         (cons (string->canonical-symbol (car x))
               (apply make-proc-obj (car x) #t #f (cdr x))))
       prim-procs))

(define (targ-prim-info name)
  (let ((x (assq name targ-prim-proc-table)))
    (if x (cdr x) #f)))

(define (targ-get-prim-info name)
  (let ((proc (targ-prim-info (string->canonical-symbol name))))
    (if proc
      proc
      (compiler-internal-error
        "targ-get-prim-info, unknown primitive:" name))))

;------------------------------------------------------------------------------
;
; Dumping of a compilation module

(define (targ-dump procs filename c-intf options)
  (targ-heap-begin!)
  (targ-use-prc (car procs) #f)
  (for-each (lambda (proc) (targ-use-prc proc #t)) procs)
  (targ-heap-dump (proc-obj-name (car procs))
                  (string-append filename ".c")
                  c-intf)
  (targ-heap-end!))

;------------------------------------------------------------------------------
;
; Object management

; These routines are used to keep track of all the Scheme objects and
; global variables that the compilation module uses.  These objects and
; variables will have to be created before the module is run (either
; at compile time, link time or run time).

(define targ-glo-vars #f)  ; global variables used
(define targ-prc-objs #f)  ; procedure objects
(define targ-lbl-alloc #f) ; label table allocation pointer
(define targ-cns-objs #f)  ; pair objects
(define targ-cns-alloc #f) ; pair object allocation pointer
(define targ-sym-objs #f)  ; interned symbol objects
(define targ-sym-alloc #f) ; interned symbol object allocation pointer
(define targ-flo-objs #f)  ; floating point numbers
(define targ-flo-alloc #f) ; floating point number allocation pointer
(define targ-sub-objs #f)  ; objects that have a subtype (vector, ...)
(define targ-sub-alloc #f) ; allocation pointer for objects with a subtype
(define targ-prc-objs* #f) ; procedure objects remaining to scan

(define targ-module-decl            #f) ; declarations from "##c-decl" calls
(define targ-module-rd-res          #f) ; set of resources read from
(define targ-module-wr-res          #f) ; set of resources written to
(define targ-module-any-closed-lbl? #f) ; true if module has closed labels

(define (targ-heap-begin!)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))
    (set! targ-glo-vars  (queue-empty))
    (set! targ-prc-objs  (queue-empty))
    (set! targ-lbl-alloc 0)
    (set! targ-cns-objs  (queue-empty))
    (set! targ-cns-alloc 0)
    (set! targ-sym-objs  (queue-empty))
    (set! targ-sym-alloc 0)
    (set! targ-flo-objs  (queue-empty))
    (set! targ-flo-alloc 0)
    (set! targ-sub-objs  (queue-empty))
    (set! targ-sub-alloc 0)
    (set! targ-prc-objs* (queue-empty))
    (set! targ-module-decl            (queue-empty))
    (set! targ-module-rd-res          (make-vector nb-res #f))
    (set! targ-module-wr-res          (make-vector nb-res #f))
    (set! targ-module-any-closed-lbl? #f)
    #f))

(define (targ-heap-end!)
  (set! targ-glo-vars #f)
  (set! targ-prc-objs #f)
  (set! targ-lbl-alloc #f)
  (set! targ-cns-objs #f)
  (set! targ-cns-alloc #f)
  (set! targ-sym-objs #f)
  (set! targ-sym-alloc #f)
  (set! targ-flo-objs #f)
  (set! targ-flo-alloc #f)
  (set! targ-sub-objs #f)
  (set! targ-sub-alloc #f)
  (set! targ-prc-objs* #f)
  (set! targ-module-decl #f)
  (set! targ-module-rd-res #f)
  (set! targ-module-wr-res #f)
  (set! targ-module-any-closed-lbl? #f)
  #f)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-use-glo glo supply?)
  (let ((x (assq glo (queue->list targ-glo-vars))))
    (if x
      (begin
        (if supply? (set-cdr! (cddr x) #t) (set-car! (cddr x) #t))
        (cadr x))
      (let ((y (targ-make-cell #f)))
        (queue-put! targ-glo-vars
                    (cons glo (cons y (if supply? (cons #f #t) (cons #t #f)))))
        y))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-use-obj obj)
  (let ((type (targ-obj-type obj)))
    (case type
      ((PROCEDURE)
       (targ-use-prc obj #f))
      ((PAIR)
       (let ((x (assq obj (queue->list targ-cns-objs))))
         (if x
           (cdr x)
           (let ((y (list "CNS" targ-cns-alloc)))
             (queue-put! targ-cns-objs (cons obj y))
             (set! targ-cns-alloc (+ targ-cns-alloc 1))
             (targ-use-obj (car obj))
             (targ-use-obj (cdr obj))
             y))))
      ((SUBTYPED)
       (let ((subtype (targ-obj-subtype obj)))
         (cond ((eq? subtype 'SYMBOL)
                (let ((x (assq obj (queue->list targ-sym-objs))))
                  (if x
                    (cdr x)
                    (let ((y (list "SYM" targ-sym-alloc)))
                      (queue-put! targ-sym-objs (cons obj y))
                      (set! targ-sym-alloc (+ targ-sym-alloc 1))
                      y))))
               (else
                (let ((x (assq obj (queue->list targ-sub-objs))))
                  (if x
                    (cdr x)
                    (let ((y (list "SUB" targ-sub-alloc)))
                      (queue-put! targ-sub-objs (cons obj y))
                      (set! targ-sub-alloc (+ targ-sub-alloc 1))
                      (case subtype
                        ((VECTOR)
                         (for-each targ-use-obj (vector->lst obj)))
                        ((RATNUM)
                         (targ-use-obj (numerator obj))
                         (targ-use-obj (denominator obj)))
                        ((CPXNUM)
                         (targ-use-obj (real-part obj))
                         (targ-use-obj (imag-part obj))))
                      y)))))))
      ((FLONUM)
       (let ((x (assq obj (queue->list targ-flo-objs))))
         (if x
           (cdr x)
           (let ((y (list "FLO" targ-flo-alloc)))
             (queue-put! targ-flo-objs (cons obj y))
             (set! targ-flo-alloc (+ targ-flo-alloc 1))
             y))))
      ((BOOLEAN)
       (if (eq? obj #t) '("TRU") '("FAL")))
      ((NULL)
       '("NUL"))
      ((UNDEF)
       '("UND"))
      ((FIXNUM)
       (list "FIX" (targ-c-long obj)))
      ((CHAR)
       (list "CHR" (targ-c-char obj)))
      (else
       (compiler-internal-error
         "targ-use-obj, unknown object 'obj'" obj)))))

(define (targ-use-prc proc supply?)
  (let* ((name (proc-obj-name proc))
         (p (if (proc-obj-code proc)
              (let ((x (assq proc (queue->list targ-prc-objs))))
                (if x
                  (cddr x)
                  (let* ((y (list "PRC" (targ-make-cell #f)))
                         (z (cons proc (cons #f y))))
                    (queue-put! targ-prc-objs z)
                    (queue-put! targ-prc-objs* z)
                    y)))
              #f)))
    (if (proc-obj-primitive? proc)
      (let ((i (targ-use-glo (string->symbol name) supply?)))
        (if p
          p
          (list "GLO" i (targ-c-id-glo name))))
      p)))

(define (targ-heap-ref-obj obj)

  (define (err)
    (compiler-internal-error
      "targ-heap-ref-obj, object was not scanned 'obj'" obj))

  (let ((type (targ-obj-type obj)))
    (case type
      ((PAIR)
       (let ((x (assq obj (queue->list targ-cns-objs))))
         (if x
           (cons "REF_CNS" (cddr x))
           (err))))
      ((SUBTYPED)
       (let ((subtype (targ-obj-subtype obj)))
         (cond ((eq? subtype 'SYMBOL)
                (let ((x (assq obj (queue->list targ-sym-objs))))
                  (if x
                    (cons "REF_SYM" (cddr x))
                    (err))))
               (else
                (let ((x (assq obj (queue->list targ-sub-objs))))
                  (if x
                    (cons "REF_SUB" (cddr x))
                    (err)))))))
      ((FLONUM)
       (let ((x (assq obj (queue->list targ-flo-objs))))
         (if x
           (cons "REF_FLO" (cddr x))
           (err))))
      ((BOOLEAN)
       (if (eq? obj #t) '("REF_TRU") '("REF_FAL")))
      ((NULL)
       '("REF_NUL"))
      ((UNDEF)
       '("REF_UND"))
      ((FIXNUM)
       (list "REF_FIX" obj))
      ((CHAR)
       (list "REF_CHR" (targ-c-char obj)))
      (else
       (compiler-internal-error
         "targ-heap-ref-obj, unknown object 'obj'" obj)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-obj-type obj)
  (cond ((false-object? obj)
         'BOOLEAN)
        ((eq? obj #t)
         'BOOLEAN)
        ((null? obj)
         'NULL)
        ((undef-object? obj)
         'UNDEF)
        ((symbol-object? obj)
         'SUBTYPED)
        ((proc-obj? obj)
         'PROCEDURE)
        ((pair? obj)
         'PAIR)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj)
                     (>= obj targ-min-fixnum) (<= obj targ-max-fixnum))
                'FIXNUM)
               ((and (inexact? (real-part obj))
                     (zero? (imag-part obj)) (exact? (imag-part obj)))
                'FLONUM)
               (else
                'SUBTYPED)))
        ((char? obj)
         'CHAR)
        (else
         'SUBTYPED)))

(define (targ-obj-subtype obj)
  (cond ((symbol-object? obj)
         'SYMBOL)
        ((number? obj)
         (cond ((and (integer? obj) (exact? obj))
                'BIGNUM)
               ((and (rational? obj) (exact? obj))
                'RATNUM)
               (else
                'CPXNUM)))
        ((vector? obj)
         'VECTOR)
        ((string? obj)
         'STRING)
        (else
         (compiler-internal-error
           "targ-obj-subtype, unknown object 'obj'" obj))))

(define (targ-eq-testable-object? obj)
  (and (not (eq? obj undef-object)) ; the undef-object denotes a non-constant
       (or (symbol-object? obj)
           (memq (targ-obj-type obj)
                 '(BOOLEAN NULL FIXNUM CHAR)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; FLONUM representation

(define (targ-flonum->inexact-exponential-format x)

  (define (exp-form-pos x y i)
    (let ((i*2 (+ i i)))
      (let ((z (if (and (not (< targ-flonum-e-bias i*2))
                        (not (< x y)))
                 (exp-form-pos x (* y y) i*2)
                 (cons x 0))))
        (let ((a (car z)) (b (cdr z)))
          (let ((i+b (+ i b)))
            (if (and (not (< targ-flonum-e-bias i+b))
                     (not (< a y)))
              (begin
                (set-car! z (/ a y))
                (set-cdr! z i+b)))
            z)))))

  (define (exp-form-neg x y i)
    (let ((i*2 (+ i i)))
      (let ((z (if (and (< i*2 targ-flonum-e-bias-minus-1)
                        (< x y))
                 (exp-form-neg x (* y y) i*2)
                 (cons x 0))))
        (let ((a (car z)) (b (cdr z)))
          (let ((i+b (+ i b)))
            (if (and (< i+b targ-flonum-e-bias-minus-1)
                     (< a y))
              (begin
                (set-car! z (/ a y))
                (set-cdr! z i+b)))
            z)))))

  (define (exp-form x)
    (if (< x targ-inexact-+1)
      (let ((z (exp-form-neg x targ-inexact-+1/2 1)))
        (set-car! z (* targ-inexact-+2 (car z)))
        (set-cdr! z (- -1 (cdr z)))
        z)
      (exp-form-pos x targ-inexact-+2 1)))

  (if (negative? x)
    (let ((z (exp-form (- targ-inexact-0 x))))
      (set-car! z (- targ-inexact-0 (car z)))
      z)
    (exp-form x)))

(define (targ-flonum->exact-exponential-format x)
  (let ((z (targ-flonum->inexact-exponential-format x)))
    (let ((y (car z)))
      (cond ((not (< y targ-inexact-+2))
             (set-car! z targ-flonum-+m-min)
             (set-cdr! z targ-flonum-e-bias-plus-1))
            ((not (< targ-inexact--2 y))
             (set-car! z targ-flonum--m-min)
             (set-cdr! z targ-flonum-e-bias-plus-1))
            (else
             (set-car! z
               (truncate (inexact->exact (* (car z) targ-inexact-m-min))))))
      (set-cdr! z (- (cdr z) targ-flonum-m-bits))
      z)))

(define (targ-flonum->bits x)

  (define (bits a b)
    (if (< a targ-flonum-+m-min)
      a
      (+ (- a targ-flonum-+m-min)
         (* (+ (+ b targ-flonum-m-bits) targ-flonum-e-bias)
            targ-flonum-+m-min))))

  (let ((z (targ-flonum->exact-exponential-format x)))
    (let ((a (car z)) (b (cdr z)))
      (if (negative? a)
        (+ targ-flonum-sign-bit (bits (- 0 a) b))
        (bits a b)))))

(define targ-flonum-m-bits         52)
(define targ-flonum-e-bits         11)
(define targ-flonum-sign-bit       #x8000000000000000) ; (expt 2 (+ targ-flonum-e-bits targ-flonum-m-bits))
(define targ-flonum-+m-min         4503599627370496)   ; (expt 2 targ-flonum-m-bits)
(define targ-flonum--m-min         -4503599627370496)  ; (- targ-flonum-+m-min)
(define targ-flonum-e-bias         1023) ; (- (expt 2 (- targ-flonum-e-bits 1)) 1)
(define targ-flonum-e-bias-plus-1  1024) ; (+ targ-flonum-e-bias 1)
(define targ-flonum-e-bias-minus-1 1022) ; (- targ-flonum-e-bias 1)

(define targ-inexact-m-min         (exact->inexact targ-flonum-+m-min))
(define targ-inexact-+2            (exact->inexact 2))
(define targ-inexact--2            (exact->inexact -2))
(define targ-inexact-+1            (exact->inexact 1))
(define targ-inexact-+1/2          (exact->inexact (/ 1 2)))
(define targ-inexact-0             (exact->inexact 0))

;------------------------------------------------------------------------------
;
; Dumping of objects

; These routines write out the C code that represents the Scheme objects
; contained in the compilation module.

(define (targ-heap-dump module-name filename c-intf)

  (let ((c-decls (c-intf-decls c-intf))
        (c-procs (c-intf-procs c-intf))
        (c-inits (c-intf-inits c-intf)))

    (c-intf-decls-set! c-intf '()) ; clear c-intf structure so front-end
    (c-intf-procs-set! c-intf '()) ; will not generate C-interface file.
    (c-intf-inits-set! c-intf '())

    ; scan and convert each procedure

    (if targ-info-port
      (begin
        (display "Dumping:" targ-info-port)
        (newline targ-info-port)))

    (let loop ()
      (if (not (queue-empty? targ-prc-objs*))
        (begin
          (targ-scan-procedure (queue-get! targ-prc-objs*))
          (loop))))

    (if targ-info-port
      (newline targ-info-port))

    ; write out C code

    (let* ((glo-vars (queue->list targ-glo-vars))
           (sym-list (queue->list targ-sym-objs))
           (cns-list (queue->list targ-cns-objs))
           (flo-list (queue->list targ-flo-objs))
           (sub-list (queue->list targ-sub-objs))
           (prc-list (queue->list targ-prc-objs))
           (prm-list (targ-extract-primitives prc-list)))

      (set! targ-port (open-output-file filename))

      (targ-dump-header module-name filename)

      (targ-define-count "GLO_COUNT" (length glo-vars))
      (targ-define-count "SYM_COUNT" (length sym-list))
      (targ-define-count "CNS_COUNT" (length cns-list))
      (targ-define-count "FLO_COUNT" (length flo-list))
      (targ-define-count "LBL_COUNT" targ-lbl-alloc)
      (targ-define-count "SUB_COUNT" targ-sub-alloc)

      (targ-display "#include \"gambit.h\"") (targ-line)

      (targ-dump-linking-info glo-vars sym-list)

      (targ-dump-sym sym-list)
      (targ-dump-cns cns-list)
      (targ-dump-flo flo-list)
      (targ-dump-sub sub-list)
      (targ-dump-prc prc-list c-decls)
      (targ-dump-ini prm-list c-inits)

      (close-output-port targ-port))))

(define (targ-dump-header module-name filename)

  (targ-display "/* file: \"")
  (targ-display filename)
  (targ-display "\", produced by Gambit-C ")
  (targ-display compiler-version)
  (targ-display " */")
  (targ-line)

  (targ-display "#define ")
  (targ-code '("MODULE_NAME "))
  (targ-display "\"") ; ********* use targ-c-string ?
  (targ-display module-name)
  (targ-display "\"")
  (targ-line)

  (targ-display "#define ")
  (targ-code '("MODULE_LINKER "))
  (targ-code (targ-c-id-linker module-name))
  (targ-line)

  (targ-display "#define ")
  (targ-code '("VERSION "))
  (targ-display header-version)
  (targ-line)

  (targ-display "#define ")
  (targ-code '("MH_PROC "))
  (targ-code (targ-c-id-host module-name))
  (targ-line))

(define (targ-dump-linking-info glo-vars sym-objs)

  (targ-code* '("BEGIN_MODULE"))

  (for-each
    (lambda (x)
      (let ((name (symbol->string (car x)))
            (demand? (caddr x))
            (supply? (cdddr x)))
        (if demand?
          (targ-code* (list "DEMAND_GLO" (targ-c-id-glo name))))
        (if supply?
          (targ-code* (list "SUPPLY_GLO" (targ-c-id-glo name))))))
    glo-vars)

  (for-each
    (lambda (x)
      (let ((name (symbol->string (car x))))
        (targ-code* (list "DEMAND_SYM" (targ-c-id-sym name)))))
    sym-objs)

  (targ-line)

  (targ-dump-section "BEGIN_GLO" "END_GLO" #f glo-vars
    (lambda (i x)
      (targ-cell-set! (cadr x) i)
      (let ((name (symbol->string (car x))))
        (targ-code* (list "DEF_GLO" (targ-c-string name)))))))

(define (targ-define-count name n)
  (if (not (= n 0))
    (begin
      (targ-display "#define ")
      (targ-code (list name))
      (targ-display " ")
      (targ-display n)
      (targ-line))))

(define (targ-define-prefix var)
  (targ-display "#undef ")
  (targ-code var)
  (targ-line)
  (targ-display "#define ")
  (targ-code var))

(define (targ-dump-section begin-name end-name comma? lst dump)
  (if (not (null? lst))
    (begin
      (targ-code* (list begin-name))
      (if comma? (targ-display " "))
      (dump 0 (car lst))
      (let loop ((i 1) (lst (cdr lst)))
        (if (not (null? lst))
          (begin
            (if comma? (targ-display ","))
            (dump i (car lst))
            (loop (+ i 1) (cdr lst)))))
      (targ-code* (list end-name))
      (targ-line))))

(define (targ-dump-sym objs)
  (targ-dump-section "BEGIN_SYM" "END_SYM" #f objs
    (lambda (i x)
      (let ((name (symbol->string (car x))))
        (targ-code* (list "DEF_SYM"
                          (targ-c-id-sym name)
                          (targ-c-string name)))))))

(define (targ-dump-cns objs)
  (targ-dump-section "BEGIN_CNS" "END_CNS" #t objs
    (lambda (i x)
      (let ((obj (car x)) (i (caddr x)))
        (targ-code* (list "DEF_CNS"
                          (targ-heap-ref-obj (car obj))
                          (targ-heap-ref-obj (cdr obj))))))))

(define (targ-dump-flo objs)
  (targ-dump-section "BEGIN_FLO" "END_FLO" #t objs
    (lambda (i x)
      (let ((obj (car x)) (i (caddr x)))
        (let* ((bits (targ-flonum->bits obj))
               (bits-hi (quotient bits #x100000000))
               (bits-lo (modulo   bits #x100000000)))
          (targ-code*
            (list "DEF_FLO"
                  (list "FLO2"
                        (targ-c-hex-ulong bits-hi)
                        (targ-c-hex-ulong bits-lo)))))))))

(define (targ-dump-sub objs)

  (define (vector-like-object def-name chunk-name id elems convert cycle)
    (let ((len (length elems)))
      (targ-code* (list def-name id len))
      (let loop1 ((i 0) (elems elems))
        (if (<= i len)
          (let ((n (min cycle (- len i))))
            (let loop2 ((j (- n 1)) (elems elems) (lst '()))
              (if (>= j 0)
                (loop2 (- j 1) (cdr elems) (cons (convert (car elems)) lst))
                (begin
                  (targ-display "               ")
                  (targ-code* (cons chunk-name (cons n (reverse lst))))
                  (if (= n cycle) (loop1 (+ i n) elems))))))))))

  (define (dump-VECTOR obj id)
    (vector-like-object "DEF_SUB_VEC" 'VEC id
                        (vector->lst obj)
                        targ-heap-ref-obj
                        1))

  (define (dump-STRING obj id)
    (vector-like-object "DEF_SUB_STR" 'STR id
                        (string->lst obj)
                        targ-c-char
                        targ-max-word-size))

  (define (dump-BIGNUM obj id)
    (vector-like-object "DEF_SUB_BIG" 'BIG id
                        (targ-BIGNUM-digits obj)
                        (lambda (x) x)
                        (quotient targ-max-word-size 2)))

  (define (dump-RATNUM obj id)
    (targ-code* (list "DEF_SUB_RAT" id
                      (targ-heap-ref-obj (numerator obj))
                      (targ-heap-ref-obj (denominator obj)))))

  (define (dump-CPXNUM obj id)
    (targ-code* (list "DEF_SUB_CPX" id
                      (targ-heap-ref-obj (real-part obj))
                      (targ-heap-ref-obj (imag-part obj)))))

  (if (not (null? objs))
    (begin
      (for-each (lambda (x)
                  (let ((obj (car x)) (id (targ-sub-name (caddr x))))
                    (let ((subtype (targ-obj-subtype obj)))
                      (case subtype
                        ((VECTOR) (dump-VECTOR obj id))
                        ((STRING) (dump-STRING obj id))
                        ((BIGNUM) (dump-BIGNUM obj id))
                        ((RATNUM) (dump-RATNUM obj id))
                        ((CPXNUM) (dump-CPXNUM obj id))
                        (else
                         (compiler-internal-error
                           "targ-dump, unknown subtype" subtype))))))
                objs)
      (targ-line)
      (targ-dump-section "BEGIN_SUB" "END_SUB" #t objs
        (lambda (i x)
          (let ((id (targ-sub-name (caddr x))))
            (targ-code* (list "DEF_SUB" id))))))))

(define (targ-dump-prc objs c-decls)

  (for-each
    (lambda (x)
      (targ-display x)
      (targ-line))
    c-decls)

  (for-each targ-code* (queue->list targ-module-decl))

  (targ-begin-cod targ-module-rd-res targ-module-wr-res
                  "BEGIN_M_COD" 'MD- 'MR- 'MW-
                  #f #f #f #f)

  (if targ-module-any-closed-lbl?
    (targ-code* '("BEGIN_M_SW_CLO"))
    (targ-code* '("BEGIN_M_SW")))

  (targ-line)

  (for-each targ-dump-cod objs)

  (if targ-module-any-closed-lbl?
    (targ-code* '("END_M_SW_CLO"))
    (targ-code* '("END_M_SW")))

  (targ-end-cod "END_M_COD")

  (targ-line)

  (targ-code* '("BEGIN_LBL"))

  (let loop1 ((i 0) (objs objs))
    (if (pair? objs)
      (let ((obj (car objs)))
        (let* ((proc (car obj))
               (p (cdr obj))
               (name (proc-obj-name proc))
               (info (car p))
               (lbls (vector-ref info 1)))
          (if (= i 0)
            (targ-display " ")
            (targ-display ","))
          (targ-code* (list "DEF_LBL_INFO"
                            (targ-c-id-host name)))
          (let loop2 ((i (+ i 1)) (lbls lbls))
            (if (pair? lbls)
              (let ((x (car lbls))
                    (host (targ-c-id-host name)))
                (targ-display ",")
                (targ-code*
                  (cons (case (vector-ref x 2)
                          ((ENTRY SUBPROC)   "DEF_LBL_PROC")
                          ((RETURN)          "DEF_LBL_RETN")
                          ((RETURN-TASK)     "DEF_LBL_RETT")
                          ((RETURN-INTERNAL) "DEF_LBL_RETI")
                          (else
                           (compiler-internal-error
                             "targ-dump-prc, unknown label type")))
                        (cons host (vector-ref x 3))))
                (loop2 (+ i 1) (cdr lbls)))
              (loop1 i (cdr objs))))))))

  (targ-code* '("END_LBL"))
  (targ-line))

(define (targ-dump-cod obj)
  (let* ((proc (car obj))
         (p (cdr obj))
         (name (proc-obj-name proc))
         (info (car p))
         (code (vector-ref info 0))
         (val-lbls (vector-ref info 1))
         (rd-res (vector-ref info 2))
         (wr-res (vector-ref info 3))
         (nb-val-lbls (length val-lbls))
         (small-switch? (<= nb-val-lbls 8))
         (any-closed-lbl? (vector-ref info 4)))

    (define (mk-glbl lbl)
      (targ-make-glbl (targ-lbl-num lbl) name))

    (define (def var val)
      (targ-define-prefix (list var))
      (targ-display " ")
      (targ-code val)
      (targ-line))

    (def "PH_PROC" (targ-c-id-host name))
    (def "PH_LBL0" (caddr p))

    (targ-begin-cod rd-res wr-res
                    "BEGIN_P_COD" 'PD- 'PR- 'PW-
                    "BEGIN_P_SLBL" "DEF_P_SLBL" "END_P_SLBL"
                    (map mk-glbl val-lbls))

    (if small-switch?
      (let ((val-lbls*
              (apply append
                     (map (lambda (lbl)
                            (list (targ-lbl-num lbl) (mk-glbl lbl)))
                          (targ-prioritize-val-lbls val-lbls)))))
        (if any-closed-lbl?
          (targ-code* (cons 'BEGIN-P-SW-CLO (cons nb-val-lbls val-lbls*)))
          (targ-code* (cons 'BEGIN-P-SW (cons nb-val-lbls val-lbls*)))))
      (if any-closed-lbl?
        (targ-code* '("BEGIN_P_SW_CLO"))
        (targ-code* '("BEGIN_P_SW"))))

    (let loop ((l code))
      (if (pair? l)
        (let ((x (car l)))
          (cond ((and (pair? x) (eq? (car x) 'LABEL))
                 (let* ((lbl (cdr x))
                        (glbl (mk-glbl lbl)))
                   (cond ((targ-lbl-val? lbl)
                          (targ-code*
                            (if small-switch?
                              (list 'DEF-SLBL
                                    nb-val-lbls (targ-lbl-num lbl) glbl)
                              (list "DEF_SLBL" (targ-lbl-num lbl) glbl))))
                         ((targ-lbl-goto? lbl)
                          (targ-code* (list "DEF_GLBL" glbl))))))
                ((and (pair? x) (eq? (car x) 'APPEND))
                 (targ-code-append (cdr x)))
                (else
                 (targ-display "   ")
                 (targ-code* x)))
          (loop (cdr l)))))

    (if small-switch?
      (if any-closed-lbl?
        (targ-code* (cons 'END-P-SW-CLO nb-val-lbls))
        (targ-code* (cons 'END-P-SW nb-val-lbls)))
      (if any-closed-lbl?
        (targ-code* '("END_P_SW_CLO"))
        (targ-code* '("END_P_SW"))))

    (targ-end-cod "END_P_COD")

    (targ-line)))

(define (targ-begin-cod rd-res wr-res beg decl rd wr
                        beg-slbl def-slbl end-slbl lbls)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (define (resource-set name1 name2 element?)
      (targ-define-prefix (cons name1 "ALL"))
      (let loop ((i 0))
        (if (< i nb-res)
          (if (element? i)
            (begin
              (targ-display " ")
              (targ-code (cons name2 (targ-res-name i)))
              (loop (+ i 1)))
            (loop (+ i 1)))))
      (targ-line))

    (resource-set
      decl
      'D-
      (lambda (i) (or (vector-ref rd-res i) (vector-ref wr-res i))))

    (resource-set
      rd
      'R-
      (lambda (i) (vector-ref rd-res i)))

    (resource-set
      wr
      'W-
      (lambda (i) (vector-ref wr-res i)))

    (targ-code* (list beg))

    (if lbls
      (begin
        (targ-code* (list beg-slbl))
        (let loop ((lst lbls))
          (if (not (null? lst))
            (let ((x (car lst)))
              (targ-code* (list def-slbl x))
              (loop (cdr lst)))))
        (targ-code* (list end-slbl))))))

(define (targ-end-cod end)
  (targ-code* (list end)))

(define (targ-dump-ini objs c-inits)

  (targ-code* '("BEGIN_INI"))

  (for-each
    (lambda (obj)
      (let* ((proc (car obj))
             (p (cdr obj))
             (name (proc-obj-name proc)))
        (targ-code* (list "DEF_PRM"
                          (targ-use-glo (string->symbol name) #f)
                          (targ-c-id-glo name)
                          (caddr p)))))
    objs)

  (for-each
    (lambda (x)
      (targ-display x)
      (targ-line))
    c-inits)

  (targ-code* '("END_INI")))

(define (targ-extract-primitives objs)
  (let loop ((objs objs) (prms '()))
    (if (pair? objs)
      (let ((obj (car objs)))
        (let ((proc (car obj)))
          (if (proc-obj-primitive? proc)
            (loop (cdr objs) (cons obj prms))
            (loop (cdr objs) prms))))
      (reverse prms))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define targ-linker-tag "")
(define targ-host-tag   "H_")
(define targ-glo-tag    "G_")
(define targ-sym-tag    "S_")

(define targ-port #f)

(define (targ-code* x)
  (targ-code x)
  (targ-line))

(define (targ-code x)

  (define (one-head head args)
    (targ-display-prefixed head)
    (targ-code-args args))

  (define (two-heads head1 head2 args)
    (targ-display-prefixed head1)
    (targ-display head2)
    (targ-code-args args))

  (define (goto-lbl n name)
    (targ-display-prefixed "L")
    (targ-code n)
    (targ-display "_")
    (targ-display (targ-name->c-id name)))

  (if (pair? x)
    (let ((head (car x)) (tail (cdr x)))
      (case head
        ((C-LONG)         (targ-display-c-long      tail))
        ((C-HEX-ULONG)    (targ-display-c-hex-ulong tail))
        ((C-CHAR)         (targ-display-c-char      tail))
        ((C-STRING)       (targ-display-c-string    tail))
        ((C-ID-LINKER)    (targ-display-c-id        targ-linker-tag tail))
        ((C-ID-HOST)      (targ-display-c-id        targ-host-tag   tail))
        ((C-ID-GLO)       (targ-display-c-id        targ-glo-tag    tail))
        ((C-ID-SYM)       (targ-display-c-id        targ-sym-tag    tail))
        ((SUB-NAME)       (targ-display-prefixed "X") (targ-code tail))
        ((GLBL)           (goto-lbl (car tail) (cadr tail)))
        ((BEGIN-P-SW)     (two-heads "BEGIN_P_SW"     (car tail) (cdr tail)))
        ((END-P-SW)       (two-heads "END_P_SW"       tail       '()))
        ((BEGIN-P-SW-CLO) (two-heads "BEGIN_P_SW_CLO" (car tail) (cdr tail)))
        ((END-P-SW-CLO)   (two-heads "END_P_SW_CLO"   tail       '()))
        ((DEF-SLBL)       (two-heads "DEF_SLBL"       (car tail) (cdr tail)))
        ((D-)             (two-heads "D_"     tail '()))
        ((R-)             (two-heads "R_"     tail '()))
        ((W-)             (two-heads "W_"     tail '()))
        ((MD-)            (two-heads "MD_"    tail '()))
        ((MR-)            (two-heads "MR_"    tail '()))
        ((MW-)            (two-heads "MW_"    tail '()))
        ((PD-)            (two-heads "PD_"    tail '()))
        ((PR-)            (two-heads "PR_"    tail '()))
        ((PW-)            (two-heads "PW_"    tail '()))
        ((PSR)            (two-heads "PSR"    tail '()))
        ((R)              (two-heads "R"      tail '()))
        ((SET-R)          (two-heads "SET_R"  (car tail) (cdr tail)))
        ((VEC)            (two-heads "VEC"    (car tail) (cdr tail)))
        ((STR)            (two-heads "STR"    (car tail) (cdr tail)))
        ((BIG)            (two-heads "BIG"    (car tail) (cdr tail)))
        ((SEQ)            (targ-code-seq tail))
        ((CELL)           (targ-code tail))
        (else             (one-head head tail))))
    (if x (targ-display x))))

(define (targ-code-args args)
  (targ-code-list
    args "," (lambda () (targ-display "(")) (lambda () (targ-display ")"))))

(define (targ-code-seq args)
  (targ-code-list args " " (lambda () 0) (lambda () 0)))

(define (targ-code-append args)
  (targ-code-list args "" (lambda () 0) (lambda () 0)))

(define (targ-code-list args sep prefix suffix)
  (if (pair? args)
    (let ((arg (car args)))
      (if arg
        (begin
          (prefix)
          (targ-code arg)
          (let loop ((args (cdr args)))
            (if (pair? args)
              (let ((arg (car args)))
                (if arg
                  (begin
                    (targ-display sep)
                    (targ-code (car args))))
                (loop (cdr args)))))
          (suffix))
        (targ-code-list (cdr args) sep prefix suffix)))))

(define (targ-display x)
  (display x targ-port))

(define (targ-line)
  (newline targ-port))

(define (targ-make-cell x)
  (cons 'CELL x))

(define (targ-cell-set! x y)
  (set-cdr! x y))

(define (targ-c-long n)
  (cons 'C-LONG n))

(define (targ-c-hex-ulong n)
  (cons 'C-HEX-ULONG n))

(define (targ-c-char c)
  (cons 'C-CHAR c))

(define (targ-c-string str)
  (cons 'C-STRING str))

(define (targ-c-id-linker name)
  (cons 'C-ID-LINKER name))

(define (targ-c-id-host name)
  (cons 'C-ID-HOST name))

(define (targ-c-id-glo name)
  (cons 'C-ID-GLO name))

(define (targ-c-id-sym name)
  (cons 'C-ID-SYM name))

(define (targ-sub-name n)
  (cons 'SUB-NAME n))

(define (targ-display-prefixed str)
  (targ-display c-id-prefix) ; c-id-prefix is defined in "_parms.scm"
  (targ-display str))

(define (targ-display-c-long n)
  (targ-display n)
  (targ-display "L"))

(define (targ-display-c-hex-ulong n)
  (targ-display "0x")
  (targ-display (number->string n 16))
  (targ-display "L"))

(define (targ-display-c-char c)
  (targ-display "'")
  (targ-display (targ-char->c-char c))
  (targ-display "'"))

(define (targ-display-c-string str)
  (targ-display "\"")
  (targ-display str) ;********** check for " and \ and \n, etc
  (targ-display "\""))

(define (targ-display-c-id type name)
  (targ-display-prefixed type)
  (targ-display (targ-name->c-id name)))

(define (targ-char->c-char c)

  (define oct-table "01234567")

  (cond ((char=? c #\\)
         "\\\\")
        ((char=? c #\')
         "\\'")
        ((char=? c #\newline)
         "\\n")
        (else
         (let ((n (character-encoding c)))
           (if (or (< n 32) (>= n 127))
             (string #\\
                     (string-ref oct-table (quotient n 64))
                     (string-ref oct-table (modulo (quotient n 8) 8))
                     (string-ref oct-table (modulo n 8)))
             (string c))))))

(define (targ-name->c-id s)
  (scheme-id->c-id s))

(define (targ-BIGNUM-digits obj)

  (define (integer->digits n)
    (if (= n 0)
      '()
      (cons (remainder n targ-bignum-radix)
            (integer->digits (quotient n targ-bignum-radix)))))

  (if (< obj 0)
    (cons 0 (integer->digits (- obj)))
    (cons 1 (integer->digits obj))))

;------------------------------------------------------------------------------

(define (targ-scan-procedure obj)
  (let* ((proc (car obj))
         (p (cdr obj))
         (nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (if targ-info-port
      (begin
        (display "  #[" targ-info-port)
        (if (proc-obj-primitive? proc)
          (display "primitive " targ-info-port)
          (display "procedure " targ-info-port))
        (display (proc-obj-name proc) targ-info-port)
        (display "]" targ-info-port)))

    (set! targ-proc-name               (proc-obj-name proc))
    (set! targ-proc-code               (queue-empty))
    (set! targ-proc-rd-res             (make-vector nb-res #f))
    (set! targ-proc-wr-res             (make-vector nb-res #f))
    (set! targ-proc-any-closed-lbl?    #f)
    (set! targ-proc-lbl-tbl            (queue-empty))
    (set! targ-proc-lbl-tbl-ord        (queue-empty))
    (set! targ-var-descr-queue         (queue-empty))
    (set! targ-first-class-label-queue (queue-empty))

    (let ((x (proc-obj-code proc)))
      (if (bbs? x)
        (targ-scan-scheme-procedure x)
        (targ-scan-c-procedure x)))

    (targ-cell-set! (caddr p) (+ targ-lbl-alloc 1))

    ; Assign label numbers sequentially, starting with "value" labels
    ; and then "goto" labels

    (let ((code (queue->list targ-proc-code))
          (ord-lbls (queue->list targ-proc-lbl-tbl-ord)))
      (let loop2 ((l ord-lbls) (i 0) (val-lbls '()))
        (if (pair? l)
          (let ((x (car l)))
            (if (targ-lbl-val? x)
              (begin
                (targ-cell-set! (targ-lbl-num x) i)
                (loop2 (cdr l) (+ i 1) (cons x val-lbls)))
              (loop2 (cdr l) i val-lbls)))
          (begin
            (set! targ-lbl-alloc (+ targ-lbl-alloc (+ i 1)))
            (set-car! p
              (vector code
                      (reverse val-lbls)
                      targ-proc-rd-res
                      targ-proc-wr-res
                      targ-proc-any-closed-lbl?))
            (let loop3 ((l ord-lbls) (i i))
              (if (pair? l)
                (let ((x (car l)))
                  (if (and (targ-lbl-goto? x)
                           (not (targ-lbl-val? x)))
                    (begin
                      (targ-cell-set! (targ-lbl-num x) i)
                      (loop3 (cdr l) (+ i 1)))
                    (loop3 (cdr l) i)))))))))

    (if targ-info-port
      (newline targ-info-port))

    ))

(define (targ-scan-scheme-procedure bbs)

  (set! targ-proc-entry-lbl   (bbs-entry-lbl-num bbs))
  (set! targ-proc-lbl-counter (bbs-lbl-counter bbs))

    (let loop ((prev-bb #f)
               (prev-gvm-instr #f)
               (l (bbs->code-list bbs)))
      (if (not (null? l))
        (let ((pres-bb (code-bb (car l)))
              (pres-gvm-instr (code-gvm-instr (car l)))
              (pres-slots-needed (code-slots-needed (car l)))
              (next-gvm-instr (if (null? (cdr l))
                                #f
                                (code-gvm-instr (cadr l)))))

          (targ-gen-gvm-instr prev-gvm-instr
                              pres-gvm-instr
                              next-gvm-instr
                              pres-slots-needed)

          (loop pres-bb pres-gvm-instr (cdr l))))))

(define (targ-scan-c-procedure c-proc)

  (define (ps-opnd opnd)
    (cond ((reg? opnd)
           (let ((n (reg-num opnd)))
             (cons 'PSR n)))
          (else ; must be stack slot
           (let ((fp-offs (- (stk-num opnd) targ-proc-fp))
                 (sp-offs (- (stk-num opnd) targ-proc-sp)))
             (list "PSSTK" fp-offs sp-offs)))))

  (let* ((arity (c-proc-arity c-proc))
         (pc (targ-label-info arity arity #f #f))
         (pc-map (pcontext-map pc))
         (pc-fs (pcontext-fs pc))
         (ret (cdr (assv 'return pc-map))))

    (set! targ-proc-entry-lbl 1)

    (targ-start-bb pc-fs)

    (targ-emit-label-entry targ-proc-entry-lbl arity #f)
    (targ-ref-lbl-goto targ-proc-entry-lbl)

    (targ-emit (list "IF_NARGS_EQ" arity '("NOTHING")))
    (targ-emit (list "WRONG_NARGS" (targ-ref-lbl-val targ-proc-entry-lbl)))
    (targ-emit (list 'APPEND
                     (list "DEF_GLBL" (targ-make-glbl "" targ-proc-name))
                     #\newline
                     '("BEGIN_C_COD")
                     #\newline))

    (let loop ((i 1))
      (if (<= i arity)
        (let ((opnd (cdr (assv i pc-map))))
          (targ-emit
            (list 'APPEND
                  "#undef " c-id-prefix "ARG" i #\newline
                  "#define " c-id-prefix "ARG" i
                  " " (ps-opnd opnd) #\newline))
          (loop (+ i 1)))))

    (targ-emit
      (list 'APPEND
            "#undef " c-id-prefix "RESULT" #\newline
            "#define " c-id-prefix "RESULT " (ps-opnd (make-reg 1)) #\newline
            (c-proc-body c-proc)
            '("END_C_COD") #\newline))

    (targ-emit
      (list 'SEQ
            (if (> pc-fs 0)
              (list "PSADJFPSP" (- pc-fs) (- pc-fs))
              #f)
            (list "JUMP"
                  '("NOTHING")
                  (ps-opnd ret))))))

;------------------------------------------------------------------------------

; Information attached to a procedure

(define targ-proc-name            #f) ; procedure's name
(define targ-proc-code            #f) ; code of the procedure
(define targ-proc-entry-lbl       #f) ; entry label
(define targ-proc-lbl-counter     #f) ; label counter
(define targ-proc-rd-res          #f) ; set of resources read from
(define targ-proc-wr-res          #f) ; set of resources written to
(define targ-proc-any-closed-lbl? #f) ; true if procedure has closed labels
(define targ-proc-lbl-tbl         #f) ; table of all labels
(define targ-proc-lbl-tbl-ord     #f) ; table of all labels ordered by def time
(define targ-proc-fp              #f) ; frame pointer
(define targ-proc-sp              #f) ; stack pointer
(define targ-proc-hp              #f) ; heap pointer

(define targ-proc-instr-source    #f)
(define targ-proc-entry-frame     #f)
(define targ-proc-exit-frame      #f)

; Emit a piece of code

(define (targ-emit code)
  (queue-put! targ-proc-code code))

; Emit a label

(define (targ-emit-label-simp lbl)
  (targ-emit-label lbl #f))

(define (targ-emit-label-entry lbl nb-parms label-descr)
  (targ-emit-label lbl 'ENTRY nb-parms 0))

(define (targ-emit-label-subproc lbl nb-parms nb-closed label-descr)
  (targ-emit-label lbl 'SUBPROC nb-parms nb-closed))

(define (targ-emit-label-return lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN fs link))

(define (targ-emit-label-return-task lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN-TASK fs link))

(define (targ-emit-label-return-internal lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN-INTERNAL fs link))

; Add a label to the code stream

(define (targ-emit-label lbl class . info)
  (let ((x (targ-get-lbl lbl)))
    (targ-emit (cons 'LABEL x))
    (targ-add-label x class info)))

(define (targ-add-label lbl-struct class info)
  (vector-set! lbl-struct 2 class)
  (vector-set! lbl-struct 3 info)
  (queue-put! targ-proc-lbl-tbl-ord lbl-struct))

; Add label "n" to label table and return label object

(define (targ-get-lbl n)
  (let ((x (assq n (queue->list targ-proc-lbl-tbl))))
    (if x
      (cdr x)
      (let ((y (vector (targ-make-cell #f) ; eventual label number (set later)
                       #f                  ; used as a "goto" label?
                       #f                  ; class (not #f if "value" label)
                       #f)))               ; extra info if "value" label
        (queue-put! targ-proc-lbl-tbl (cons n y))
        y))))

(define (targ-lbl-num lbl-struct)
  (vector-ref lbl-struct 0))

; Mark a label as referenced for "value" and return eventual label number

(define (targ-ref-lbl-val n)
  (let ((x (targ-get-lbl n)))
    (targ-lbl-num x)))

(define (targ-lbl-val? lbl-struct)
  (vector-ref lbl-struct 2))

; Mark a label as target for "goto" and return eventual label number

(define (targ-ref-lbl-goto n)
  (let ((x (targ-get-lbl n)))
    (vector-set! x 1 #t)
    (targ-make-glbl (targ-lbl-num x) targ-proc-name)))

(define (targ-lbl-goto? lbl-struct)
  (vector-ref lbl-struct 1))

(define (targ-make-glbl n name)
  (list 'GLBL n name))

; To generate new, unique labels

(define (targ-new-lbl)
  (targ-proc-lbl-counter))

; Heuristic to sort "value" labels in decreasing order of importance

(define (targ-prioritize-val-lbls lbls)

  (define (extract l class)
    (let loop ((l l) (r '()))
      (cond ((null? l)
             r)
            ((eq? class (vector-ref (car l) 2))
             (loop (cdr l) (cons (car l) r)))
            (else
             (loop (cdr l) r)))))

  (append (extract lbls 'RETURN)
          (extract lbls 'RETURN-TASK)
          (extract lbls 'SUBPROC)
          (extract lbls 'RETURN-INTERNAL)
          (extract lbls 'ENTRY)))

(define (targ-heap-reserve space fs)
  (set! targ-proc-hp (+ targ-proc-hp space))
  (if (> targ-proc-hp targ-heap-fudge)
    (targ-check-heap space fs)))

(define (targ-check-heap space fs)
  (let ((lbl (targ-new-lbl)))
    (targ-need-heap)
    (targ-emit (list 'SEQ
                     (targ-adjust-stack fs)
                     (list "CHECK_HEAP"
                           (targ-ref-lbl-val lbl)
                           (+ targ-heap-fudge space))))
    (targ-gen-label-return* lbl 'RETURN-INTERNAL fs)
    (set! targ-proc-hp 0)))

(define (targ-poll fs)
  (let ((lbl (targ-new-lbl)))
    (targ-rd-sp)
    (targ-emit (list 'SEQ
                     (targ-adjust-stack fs)
                     (list "POLL" (targ-ref-lbl-val lbl))))
    (targ-gen-label-return* lbl 'RETURN-INTERNAL fs)))

(define (targ-start-bb fs)
  (set! targ-proc-hp 0)
  (set! targ-proc-fp fs)
  (set! targ-proc-sp fs))

; Management of resources

(define targ-nb-non-reg-res 2)

(define (targ-res-name i)
  (if (< i targ-nb-non-reg-res)
    (vector-ref '#("HEAP" "SP") i)
    (string-append "R" (number->string (- i targ-nb-non-reg-res)))))

(define (targ-rd-res i)
  (vector-set! targ-module-rd-res i #t)
  (vector-set! targ-proc-rd-res i #t))

(define (targ-wr-res i)
  (targ-rd-res i)
  (vector-set! targ-module-wr-res i #t)
  (vector-set! targ-proc-wr-res i #t))

(define (targ-need-heap)
  (targ-wr-res 0))

(define (targ-rd-sp)
  (targ-rd-res 1))

(define (targ-wr-sp)
  (targ-wr-res 1))

(define (targ-rd-reg n)
  (targ-rd-res (+ n targ-nb-non-reg-res)))

(define (targ-wr-reg n)
  (targ-wr-res (+ n targ-nb-non-reg-res)))

(define (targ-need-closed)
  (set! targ-module-any-closed-lbl? #t)
  (set! targ-proc-any-closed-lbl? #t)
  (targ-wr-reg (+ targ-nb-arg-regs 1)))

(define (targ-use-all-res)
  (let loop ((i (- (+ targ-nb-non-reg-res targ-nb-gvm-regs) 1)))
    (if (>= i 0)
      (begin
        (targ-wr-res i)
        (loop (- i 1))))))

;------------------------------------------------------------------------------

(define (targ-add-var-descr! descr)

  (define (index x l)
    (let loop ((l l) (i 0))
      (cond ((not (pair? l))    #f)
            ((equal? (car l) x) i)
            (else               (loop (cdr l) (+ i 1))))))

  (let ((n (index descr (queue->list targ-var-descr-queue))))
    (if n
      n
      (let ((m (length (queue->list targ-var-descr-queue))))
        (queue-put! targ-var-descr-queue descr)
        m))))

(define (targ-add-first-class-label! source slots frame)
  (let loop ((i 0) (l1 slots) (l2 '()))
    (if (pair? l1)
      (let ((var (car l1)))
        (let ((x (frame-live? var frame)))
          (if (and x (or (pair? x) (not (temp-var? x))))
            (let ((descr-index
                    (targ-add-var-descr!
                     (if (pair? x)
                        (map (lambda (y) (targ-add-var-descr! (var-name y))) x)
                        (var-name x)))))
              (loop (+ i 1) (cdr l1) (cons (+ (* i 16384) descr-index) l2)))
            (loop (+ i 1) (cdr l1) l2))))
      (let ((label-descr (lst->vector (cons 0 (cons source l2)))))
        (queue-put! targ-first-class-label-queue label-descr)
        label-descr))))

;------------------------------------------------------------------------------

(define (targ-gen-gvm-instr prev-gvm-instr gvm-instr next-gvm-instr sn)

  (set! targ-proc-instr-source (comment-get (gvm-instr-comment gvm-instr) 'SOURCE))
  (set! targ-proc-exit-frame   (gvm-instr-frame gvm-instr))
  (set! targ-proc-entry-frame  (and prev-gvm-instr (gvm-instr-frame prev-gvm-instr)))

  (case (gvm-instr-type gvm-instr)

    ((LABEL)
     (set! targ-proc-entry-frame targ-proc-exit-frame)
     (targ-start-bb (frame-size targ-proc-exit-frame))
     (case (LABEL-type gvm-instr)
       ((SIMPLE)
        (targ-gen-LABEL-SIMPLE (LABEL-lbl-num gvm-instr)
                               sn))
       ((ENTRY)
        (targ-gen-LABEL-ENTRY (LABEL-lbl-num gvm-instr)
                              (LABEL-ENTRY-nb-parms gvm-instr)
                              (LABEL-ENTRY-min gvm-instr)
                              (LABEL-ENTRY-rest? gvm-instr)
                              (LABEL-ENTRY-closed? gvm-instr)
                              sn))
       ((RETURN)
        (targ-gen-LABEL-RETURN (LABEL-lbl-num gvm-instr)
                               sn))
       ((TASK-ENTRY)
        (targ-gen-LABEL-TASK-ENTRY (LABEL-lbl-num gvm-instr)
                                   sn))
       ((TASK-RETURN)
        (targ-gen-LABEL-TASK-RETURN (LABEL-lbl-num gvm-instr)
                                    sn))
       (else
        (compiler-internal-error
          "targ-gen-gvm-instr, unknown label type"))))

    ((APPLY)
     (targ-gen-APPLY (APPLY-prim gvm-instr)
                     (APPLY-opnds gvm-instr)
                     (APPLY-loc gvm-instr)
                     sn))

    ((COPY)
     (targ-gen-COPY (COPY-opnd gvm-instr)
                    (COPY-loc gvm-instr)
                    sn))

    ((CLOSE)
     (targ-gen-CLOSE (CLOSE-parms gvm-instr)
                     sn))

    ((IFJUMP)
     (targ-gen-IFJUMP (IFJUMP-test gvm-instr)
                      (IFJUMP-opnds gvm-instr)
                      (IFJUMP-true gvm-instr)
                      (IFJUMP-false gvm-instr)
                      (IFJUMP-poll? gvm-instr)
                      (if (and next-gvm-instr
                               (memq (LABEL-type next-gvm-instr)
                               '(SIMPLE TASK-ENTRY)))
                        (LABEL-lbl-num next-gvm-instr)
                        #f)))

    ((JUMP)
     (targ-gen-JUMP (JUMP-opnd gvm-instr)
                    (JUMP-nb-args gvm-instr)
                    (JUMP-poll? gvm-instr)
                    (if (and next-gvm-instr
                             (memq (LABEL-type next-gvm-instr)
                                   '(SIMPLE TASK-ENTRY)))
                      (LABEL-lbl-num next-gvm-instr)
                      #f)))

    (else
     (compiler-internal-error
       "targ-gen-gvm-instr, unknown 'gvm-instr'" gvm-instr))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-SIMPLE lbl sn)
  (targ-emit-label-simp lbl))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-ENTRY lbl nb-parms min rest? closed? sn)

  (let ((label-descr (targ-add-first-class-label!
                       targ-proc-instr-source
                       '()
                       targ-proc-exit-frame)))
    (if (= lbl targ-proc-entry-lbl)
      (begin
        (targ-emit-label-entry lbl nb-parms label-descr)
        (targ-ref-lbl-val lbl)
        (targ-ref-lbl-goto lbl))
      (let ((nb-closed (length (frame-closed targ-proc-exit-frame))))
        (targ-emit-label-subproc lbl nb-parms nb-closed label-descr))))

  (let ((nb-parms* (if rest? (- nb-parms 1) nb-parms))
        (lbl* (targ-ref-lbl-val lbl)))

    (define (check-args i)
      (if (> i nb-parms*)

        (targ-emit
          (if rest?
            (if closed? '("BUILD_REST_CLO") (list "BUILD_REST" lbl*))
            (if closed? '("WRONG_NARGS_CLO") (list "WRONG_NARGS" lbl*))))

        (let ((nb-stacked (max 0 (- i targ-nb-arg-regs)))
              (nb-stacked* (max 0 (- nb-parms targ-nb-arg-regs))))

          (define (setup-args j)
            (if (<= j nb-parms)
              (let* ((rest (setup-args (+ j 1)))
                     (src-reg (- j nb-stacked))
                     (src (cond ((<= j i) (cons 'R src-reg))
                                ((and rest? (= j nb-parms)) '("NUL"))
                                (else '("UNA")))))
                (if (<= j nb-stacked*)
                  (begin
                    (if (<= j i) (targ-rd-reg src-reg))
                    (cons (list "PUSH" src) rest))
                  (if (and (<= j i) (= nb-stacked nb-stacked*))
                    rest
                    (let ((dst-reg (- j nb-stacked*)))
                      (if (<= j i) (targ-rd-reg src-reg))
                      (targ-wr-reg dst-reg)
                      (cons (list 'SET-R dst-reg src) rest)))))
              '()))

          (let ((x (setup-args (+ nb-stacked 1))))
            (targ-emit (list "IF_NARGS_EQ"
                             i
                             (if (null? x) '("NOTHING") (cons 'SEQ x)))))

          (check-args (+ i 1)))))

    (check-args min)

    (if (= lbl targ-proc-entry-lbl)
      (targ-emit (list 'APPEND
                       (list "DEF_GLBL" (targ-make-glbl "" targ-proc-name))
                       #\newline)))

    (if closed?
      (targ-need-closed))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-RETURN lbl sn)
  (targ-gen-label-return* lbl 'RETURN (frame-size targ-proc-exit-frame)))

(define (targ-gen-label-return* lbl kind fs)

  (define (adjust-slots l n)
    (cond ((= n 0) l)
          ((< n 0) (adjust-slots (cdr l) (+ n 1)))
          (else    (adjust-slots (cons empty-var l) (- n 1)))))

  (let* ((frame targ-proc-entry-frame)
         (slots (reverse (adjust-slots (frame-slots frame)
                                       (- fs (frame-size frame)))))
         (regs (frame-regs frame))
         (vars (if (eq? kind 'RETURN-INTERNAL)
                 (append slots regs)
                 slots)))
    (let ((label-descr
            (targ-add-first-class-label!
              targ-proc-instr-source
              vars
              frame)))
      (let ((link (pos-in-list ret-var vars)))
        (if link
          (let ((n (length slots)))
            (case kind
              ((RETURN)
               (targ-emit-label-return lbl n link label-descr))
              ((RETURN-TASK)
               (targ-emit-label-return-task lbl n link label-descr))
              ((RETURN-INTERNAL)
               (targ-emit-label-return-internal lbl n link label-descr))
              (else
               (compiler-internal-error
                 "targ-gen-label-return*, unknown label kind"))))
          (compiler-internal-error
            "targ-gen-label-return*, no return address in frame"))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-TASK-ENTRY lbl sn)

  (targ-emit-label-simp lbl)

  (let ((fs (frame-size targ-proc-exit-frame)))
    (if (= fs 0)

      (targ-emit '("PUSH" (r . 0)))

      (begin
       (targ-emit '("ADJFPSP" 1 1))
        (let loop ((i 0))
          (if (< i fs)
            (begin
              (targ-emit (list "SET_STK" (- i) (- i)
                               (list "STK" (- (+ i 1)) (- (+ i 1)))))
              (loop (+ i 1)))
            (targ-emit (list "SET_STK" (- i) (- i) '(r . 0)))))))

    (targ-emit (list "TASK_PUSH" fs))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-TASK-RETURN lbl sn)
  (let ((lbl2 (targ-new-lbl))
        (fs (+ (frame-size targ-proc-exit-frame) 1)))

    (targ-start-bb fs)

    (targ-gen-label-return* lbl 'RETURN-TASK fs)
    (targ-emit (list "TASK_POP" (targ-ref-lbl-val lbl2)))
    (targ-gen-label-return* lbl2 'RETURN fs)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-APPLY prim opnds loc sn)
  (let ((proc (proc-obj-inlinable prim)))
    (if proc
      (proc opnds loc sn)
      (compiler-internal-error
        "targ-gen-APPLY, unknown 'prim'" prim))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-COPY opnd loc sn)
  (targ-emit
    (targ-loc
      loc
      sn
      (targ-uses-stk? opnd)
      (targ-opnd opnd sn (targ-uses-stk? loc)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-CLOSE parms sn)

  (define (close parms* sn*)
    (if (pair? parms*)

      (let* ((parm (car parms*))
             (lbl (closure-parms-lbl parm))
             (loc (closure-parms-loc parm))
             (opnds (closure-parms-opnds parm))
             (sn** (targ-sn-opnds opnds sn*)))
        (close (cdr parms*) (targ-sn-loc loc sn**))
        (let ((x (targ-opnd loc sn** #f)))
          (targ-emit (list "SETUP_CLO"
                           x
                           (targ-ref-lbl-val lbl)
                           (targ-stores opnds sn* "STORE")))))

      (begin

        (targ-heap-reserve
          (apply +
                 (map (lambda (parm)
                        (targ-closure-space
                          (length (closure-parms-opnds parm))))
                      parms))
          sn*)

        (for-each (lambda (parm)
                    (let ((loc (closure-parms-loc parm))
                          (opnds (closure-parms-opnds parm)))
                      (targ-emit
                        (targ-loc loc sn* #f (list "ALLOC_CLO"
                                                   (length opnds))))))
                  parms))))

  (close (reverse parms) sn))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-IFJUMP test opnds true-lbl false-lbl poll? next-lbl)
  (let ((proc (proc-obj-test test)))
    (if proc

      (let* ((fs (frame-size targ-proc-exit-frame))
             (fs* (targ-sn-opnds opnds fs)))

        (if (> targ-proc-hp 0)
          (targ-check-heap 0 fs*))

        (if poll?
          (targ-poll fs*))

        (let ((adj-stk (targ-adjust-stack fs)))

          (define (emit adj-stk test lbl other)
            (targ-emit
              (list 'SEQ
                    adj-stk
                    (list "IF_GOTO" test (targ-ref-lbl-goto lbl))
                    other)))

          (cond ((eqv? true-lbl next-lbl)
                 (emit adj-stk (proc #t opnds fs) false-lbl #f))
                ((eqv? false-lbl next-lbl)
                 (emit adj-stk (proc #f opnds fs) true-lbl #f))
                (else
                 (emit adj-stk (proc #f opnds fs) true-lbl
                       (list "GOTO" (targ-ref-lbl-goto false-lbl)))))))

      (compiler-internal-error
        "targ-gen-IFJUMP, unknown 'test'" test))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-JUMP opnd nb-args poll? next-lbl)
  (let* ((fs (frame-size targ-proc-exit-frame))
         (fs* (targ-sn-opnd opnd fs)))

    (if (> targ-proc-hp 0)
      (targ-check-heap 0 fs*))

    (if poll?
      (targ-poll fs*))

    (let* ((adj-stk (targ-adjust-stack fs))
           (set-nargs (if nb-args (list "SET_NARGS" nb-args) #f)))

      (cond ((lbl? opnd)
             (let ((n (lbl-num opnd)))
               (if (and next-lbl (= next-lbl n)) ; fall through?
                 (if (or adj-stk set-nargs)
                   (targ-emit
                     (list 'SEQ
                           adj-stk
                           set-nargs)))
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         set-nargs
                         (list "GOTO" (targ-ref-lbl-goto n)))))))
            ((and (obj? opnd)
                  (proc-obj? (obj-val opnd))
                  (proc-obj-code (obj-val opnd)))
             (let* ((proc (obj-val opnd))
                    (x (targ-use-obj proc))
                    (name (proc-obj-name proc)))
               (if (and nb-args (targ-arg-check-avoidable? proc nb-args))
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         (list "JUMPLOCAL"
                               set-nargs
                               (list "PRC" (cadr x))
                               (targ-make-glbl "" name))))
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         set-nargs
                         (list "JUMPLOCAL"
                               '("NOTHING")
                               (list "PRC" (cadr x))
                               (targ-make-glbl 0 name)))))))
            (else
             (targ-emit
               (list 'SEQ
                     adj-stk
                     (list "JUMP"
                           (if nb-args set-nargs '("NOTHING"))
                           (targ-opnd opnd fs #t)))))))))

(define (targ-arg-check-avoidable? proc nb-args)
  (let ((x (proc-obj-call-pat proc)))
    (if (and (pair? x) (null? (cdr x))) ; proc accepts a fixed nb of args?
      (let ((arg-count (car x)))
        (= arg-count nb-args))          ; nb of arguments = nb of parameters?
      #f)))

;------------------------------------------------------------------------------

(define (targ-opnd opnd sn sp-frozen?)

  (cond ((reg? opnd)
         (let ((n (reg-num opnd)))
           (targ-rd-reg n)
           (cons 'R n)))

        ((stk? opnd)
         (let ((fp-offs (- (stk-num opnd) targ-proc-fp))
               (sp-offs (- (stk-num opnd) targ-proc-sp)))
           (targ-rd-sp)
           (if (and (not sp-frozen?) (= sp-offs 0) (< sn targ-proc-sp))
             (begin
               (targ-wr-sp)
               (set! targ-proc-sp (- targ-proc-sp 1))
               (list "POP" fp-offs))
             (list "STK" fp-offs sp-offs))))

        ((glo? opnd)
         (let ((name (glo-name opnd)))
           (list "GLO"
                 (targ-use-glo name #f)
                 (targ-c-id-glo (symbol->string name)))))

        ((clo? opnd)
         (list "CLO"
               (targ-opnd (clo-base opnd) sn sp-frozen?)
               (clo-index opnd)))

        ((lbl? opnd)
         (let ((n (lbl-num opnd)))
           (list "LBL" (targ-ref-lbl-val n))))

        ((obj? opnd)
         (targ-use-obj (obj-val opnd)))

        (else
         (compiler-internal-error
           "targ-opnd, unknown 'opnd'" opnd))))

(define (targ-loc loc sn sp-frozen? val)

  (cond ((reg? loc)
         (let ((n (reg-num loc)))
           (targ-wr-reg n)
           (list 'SET-R n val)))

        ((stk? loc)
         (let ((fp-offs (- (stk-num loc) targ-proc-fp))
               (sp-offs (- (stk-num loc) targ-proc-sp)))
           (targ-rd-sp)
           (if (and (not sp-frozen?) (= sp-offs 1))
             (begin
               (targ-wr-sp)
               (set! targ-proc-sp (+ targ-proc-sp 1))
               (list "SET_PUSH" fp-offs val))
             (list "SET_STK" fp-offs sp-offs val))))

        ((glo? loc)
         (let ((name (glo-name loc)))
           (list "SET_GLO"
                 (targ-use-glo name #t)
                 (targ-c-id-glo (symbol->string name))
                 val)))

        ((clo? loc)
         (list "SET_CLO"
               (targ-opnd (clo-base loc) sn sp-frozen?)
               (clo-index loc)))

        (else
         (compiler-internal-error
           "targ-loc, unknown 'loc'" loc))))

(define (targ-stores opnds sn kind)

  (define (stores opnds sn)
    (if (pair? opnds)
      (let ((opnd (car opnds)))
        (let ((lst (stores (cdr opnds) (targ-sn-opnd opnd sn))))
          (cons (list kind (targ-opnd opnd sn #f)) lst)))
      '()))

  (cons 'SEQ (reverse (stores (reverse opnds) sn))))

(define (targ-adjust-stack fs)
  (let ((fp targ-proc-fp) (sp targ-proc-sp))
    (set! targ-proc-fp fs)
    (set! targ-proc-sp fs)
    (if (and (= fs fp) (= fs sp))
      #f
      (begin
        (targ-wr-sp)
        (cond ((= fs sp)
               (list "ADJFP" (- fs fp)))
              ((= fs fp)
               (list "ADJSP" (- fs sp)))
              (else
               (list "ADJFPSP" (- fs fp) (- fs sp))))))))

(define (targ-uses-stk? opnd)
  (cond ((stk? opnd) #t)
        ((clo? opnd) (targ-uses-stk? (clo-base opnd)))
        (else        #f)))

(define (targ-nb-use-stk l)
  (let loop ((l l) (n 0))
    (if (pair? l)
      (if (targ-uses-stk? (car l))
        (loop (cdr l) (+ n 1))
        (loop (cdr l) n))
      n)))

(define (targ-sn-opnd opnd sn)
  (cond ((stk? opnd)
         (max (stk-num opnd) sn))
        ((clo? opnd)
         (targ-sn-opnd (clo-base opnd) sn))
        (else
         sn)))

(define (targ-sn-opnds opnds sn)
  (if (pair? opnds)
    (targ-sn-opnd (car opnds) (targ-sn-opnds (cdr opnds) sn))
    sn))

(define (targ-sn-loc loc sn)
  (if loc
    (targ-sn-opnd loc sn)
    sn))

;==============================================================================

; DATABASE OF PRIMITIVES

(set! targ-prim-proc-table
  (append
    (map (lambda (x)
           (cons (string->canonical-symbol (car x))
                 (apply make-proc-obj (car x) #t #f (cdr x))))
'(
("##c-code"  0            #t 0        (#f)   )
("##c-decl"  0            #t 0        #f     )
))
targ-prim-proc-table))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedure specialization

(define (targ-spec name specializer-maker)
  (let ((proc-name (string->canonical-symbol name)))
    (let ((proc (targ-get-prim-info name)))
      (if proc
        (proc-obj-specialize-set! proc (specializer-maker proc proc-name))
        (compiler-internal-error
          "targ-spec, unknown primitive:" name)))))

; Safe specialization

(define (targ-s name)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info name)))
      (lambda (decls args) spec))))

; Unsafe specialization

(define (targ-u name)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info name)))
      (lambda (decls args) (if (not (safe? decls)) spec proc)))))

; Safe arithmetic specialization

(define (targ-s-a fix-name flo-name)
  (targ-arith #t fix-name flo-name))

; Unsafe arithmetic specialization

(define (targ-u-a fix-name flo-name)
  (targ-arith #f fix-name flo-name))

(define (targ-arith fix-safe? fix-name flo-name)
  (lambda (proc proc-name)
    (let ((fix-spec (if fix-name (targ-get-prim-info fix-name) proc))
          (flo-spec (if flo-name (targ-get-prim-info flo-name) proc)))
      (lambda (decls args)
        (let ((arith (arith-implementation proc-name decls)))
          (cond ((eq? arith FIXNUM-sym)
                 (if (or fix-safe? (not (safe? decls))) fix-spec proc))
                ((eq? arith FLONUM-sym)
                 (if (not (safe? decls)) flo-spec proc))
                (else
                 proc)))))))

; Safe specialization for ##eqv?

(define (targ-s-**eqv?)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info "##eq?")))
      (lambda (decls args)
        (if (and (= (length args) 2)
                 (or (targ-eq-testable-object? (car args))
                     (targ-eq-testable-object? (cadr args))))
          spec
          proc)))))

;------------------------------------------------------------------------------

(define (targ-op name descr)
  (descr (targ-get-prim-info name)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-alloc compute-space f)
  (targ-setup-inlinable-proc
    #f
    (lambda (opnds sn sp-frozen?)
      (targ-heap-reserve
        (compute-space (length opnds))
        (targ-sn-opnds opnds sn))
      (f opnds sn sp-frozen?))))

(define (targ-APPLY-cons)
  (targ-APPLY-alloc
    (lambda (n) targ-cons-space)
    (targ-APPLY-simp-generator "CONS")))

(define (targ-APPLY-make-cell)
  (targ-APPLY-alloc
    (lambda (n) targ-cell-space)
    (targ-APPLY-simp-generator "MAKECELL")))

(define (targ-APPLY-make-placeholder)
  (targ-APPLY-alloc
    (lambda (n) targ-placeholder-space)
    (targ-APPLY-simp-generator "MAKEPLACEHOLDER")))

(define (targ-APPLY-vector kind)

  (define (compute-space n)
    (case kind
      ((STRING)   (targ-string-space n))
      ((VECTOR8)  (targ-vector8-space n))
      ((VECTOR16) (targ-vector8-space (* n 2)))
      (else       (targ-vector-space n))))

  (define allocator-name
    (case kind
      ((STRING)   "ALLOC_STRING")
      ((VECTOR8)  "ALLOC_VECTOR8")
      ((VECTOR16) "ALLOC_VECTOR16")
      (else       "ALLOC_VECTOR")))

  (define getter-operation
    (case kind
      ((STRING)   '("GET_STRING"))
      ((VECTOR8)  '("GET_VECTOR8"))
      ((VECTOR16) '("GET_VECTOR16"))
      (else       '("GET_VECTOR"))))

  (define store-operation
    (case kind
      ((STRING)   "STORECHAR")
      ((VECTOR8)  "STOREINT")
      ((VECTOR16) "STOREINT")
      (else       "STORE")))

  (targ-APPLY-alloc
    compute-space
    (lambda (opnds sn sp-frozen?)
      (targ-emit
        (list allocator-name
              (length opnds)
              (targ-stores opnds sn store-operation)))
      getter-operation)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-touch)
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (let ((lbl (targ-new-lbl))
              (opnd (car opnds))
              (sn* (targ-sn-loc loc sn)))

          (targ-emit (list "TOUCH1"
                           (targ-ref-lbl-val lbl)
                           (targ-opnd opnd sn* #f)))
          (targ-gen-label-return* lbl 'RETURN-INTERNAL sn*)
          (targ-emit (list "TOUCH2"))
          (if loc
            (targ-emit (targ-loc loc sn #f (list "TOUCH3")))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-IFJUMP-simp name)
  (targ-setup-test-proc*
    (targ-IFJUMP-simp-generator name)))

(define (targ-IFJUMP-fold name)
  (targ-setup-test-proc*
    (targ-IFJUMP-fold-generator name)))

(define (targ-IFJUMP-APPLY name)
  (targ-setup-inlinable-proc*
    (targ-APPLY-simp-generator name)))

(define (targ-APPLY-simp side-effects? name)
  (targ-setup-inlinable-proc
    side-effects?
    (targ-APPLY-simp-generator name)))

(define (targ-APPLY-fold name0 name1 name2)
  (targ-setup-inlinable-proc
    #f
    (targ-APPLY-fold-generator name0 name1 name2 #f #f)))

(define (targ-APPLY-simpflo name) ; primitive constructs a flonum
  (targ-APPLY-alloc
    (lambda (n) targ-flonum-space)
    (targ-APPLY-simp-generator name)))

(define (targ-APPLY-foldflo name0 name1 name2) ; primitive constructs a flonum
  (targ-APPLY-alloc
    (lambda (n) targ-flonum-space)
    (targ-APPLY-fold-generator name0 name1 name2 "FLOBOX" "FLOUNBOX")))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-setup-test-proc* generator)
  (lambda (prim)
    ((targ-setup-test-proc generator)
     prim)
    ((targ-setup-inlinable-proc
       #f
       (lambda (opnds sn sp-frozen?)
         (list "BOOL" (generator opnds sn sp-frozen?))))
     prim)))

(define (targ-setup-test-proc generator)
  (lambda (prim)
    (proc-obj-test-set!
      prim
      (lambda (not? opnds fs)
        (let ((test (generator opnds fs #t)))
          (if not?
            (list "NOT" test)
            test))))))

(define (targ-IFJUMP-simp-generator name)
  (lambda (opnds fs sp-frozen?)
    (targ-IFJUMP-simp-gen opnds fs sp-frozen? name)))

(define (targ-IFJUMP-simp-gen opnds fs sp-frozen? name)
  (let ((frozen? (or sp-frozen? (> (targ-nb-use-stk opnds) 1))))
    (let loop ((l opnds) (args '()))
      (if (pair? l)
        (loop (cdr l) (cons (targ-opnd (car l) fs frozen?) args))
        (cons name (reverse args))))))

(define (targ-IFJUMP-fold-generator name)
  (lambda (opnds fs sp-frozen?)
    (targ-IFJUMP-fold-gen opnds fs sp-frozen? name)))

(define (targ-IFJUMP-fold-gen opnds fs sp-frozen? name)

  (define (multi-opnds opnds fs frozen?)
    (let* ((opnd1 (car opnds))
           (opnd1-uses-stk? (targ-uses-stk? opnd1))
           (opnd2 (cadr opnds))
           (opnd1* (targ-opnd opnd1 fs frozen?))
           (opnd2* (targ-opnd opnd2 fs (or frozen?
                                           opnd1-uses-stk?
                                           (pair? (cddr opnds)))))
           (r (list name opnd1* opnd2*)))
      (if (pair? (cddr opnds))
        (list "AND" r (multi-opnds (cdr opnds) fs #t))
        r)))

  (cond ((or (not (pair? opnds))
             (not (pair? (cdr opnds))))
         1)
        (else
         (multi-opnds opnds fs sp-frozen?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-setup-inlinable-proc* generator)
  (lambda (prim)
    ((targ-setup-test-proc
       (lambda (opnds fs sp-frozen?)
         (list "NOT" (list "FALSEP" (generator opnds fs sp-frozen?)))))
     prim)
    ((targ-setup-inlinable-proc #f generator)
     prim)))

(define (targ-setup-inlinable-proc side-effects? generator)
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (if side-effects? ; always generate code for side-effecting operations
          (let ((x (generator opnds sn #f)))
            (targ-emit ; if a result is needed, copy the first argument
              (if (and loc (not (eqv? (car opnds) loc)))
                (list 'SEQ
                      x
                      (targ-loc loc
                                sn
                                (targ-uses-stk? (car opnds))
                                (targ-opnd (car opnds) sn (targ-uses-stk? loc))))
                x)))
          (if loc ; only generate functional operations if result needed
            (let ((opnds-use-stk? (> (targ-nb-use-stk opnds) 0))
                  (loc-uses-stk? (and loc (targ-uses-stk? loc)))
                  (sn* (targ-sn-loc loc sn)))
              (let ((x (generator opnds sn* loc-uses-stk?)))
                (targ-emit (targ-loc loc sn opnds-use-stk? x))))))))))

(define (targ-APPLY-simp-generator name)
  (lambda (opnds sn sp-frozen?)
    (targ-APPLY-simp-gen opnds sn sp-frozen? name)))

(define (targ-APPLY-simp-gen opnds sn sp-frozen? name)
  (let ((frozen? (or sp-frozen? (> (targ-nb-use-stk opnds) 1))))
    (let loop ((l opnds) (args '()))
      (if (pair? l)
        (loop (cdr l) (cons (targ-opnd (car l) sn frozen?) args))
        (cons name (reverse args))))))

(define (targ-APPLY-fold-generator name0 name1 name2 name3 name4)
  (lambda (opnds sn sp-frozen?)
    (targ-APPLY-fold-gen opnds sn sp-frozen? name0 name1 name2 name3 name4)))

(define (targ-APPLY-fold-gen opnds sn sp-frozen? name0 name1 name2 name3 name4)
  (cond ((not (pair? opnds))
         (list name0))
        ((not (pair? (cdr opnds)))
         (list name1 (targ-opnd (car opnds) sn sp-frozen?)))
        (else
         (let ((frozen? (or sp-frozen? (> (targ-nb-use-stk opnds) 1))))
           (let* ((opnd1 (car opnds))
                  (opnd1* (targ-wrap name4 (targ-opnd opnd1 sn frozen?))))
             (let loop ((l (cdr opnds)) (r opnd1*))
               (if (pair? l)
                 (let ((opnd (car l)))
                   (loop (cdr l)
                         (list name2
                               r
                               (targ-wrap
                                 name4
                                 (targ-opnd opnd sn frozen?)))))
                 (targ-wrap name3 r))))))))

(define (targ-wrap name arg)
  (if name
    (list name arg)
    arg))

;------------------------------------------------------------------------------

; Table of inlinable operations (for APPLY and IFJUMP gvm instructions)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##TYPE"                (targ-APPLY-simp #f "TYPE"))
(targ-op "##TYPE-CAST"           (targ-APPLY-simp #f "TYPECAST"))
(targ-op "##SUBTYPE"             (targ-APPLY-simp #f "SUBTYPE"))
(targ-op "##SUBTYPE-SET!"        (targ-APPLY-simp #t "SUBTYPESET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##NOT"                 (targ-IFJUMP-simp "FALSEP"))
(targ-op "##NULL?"               (targ-IFJUMP-simp "NULLP"))
(targ-op "##UNASSIGNED?"         (targ-IFJUMP-simp "UNASSIGNEDP"))
(targ-op "##UNBOUND?"            (targ-IFJUMP-simp "UNBOUNDP"))
(targ-op "##EQ?"                 (targ-IFJUMP-simp "EQP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##FIXNUM?"             (targ-IFJUMP-simp "FIXNUMP"))
(targ-op "##FLONUM?"             (targ-IFJUMP-simp "FLONUMP"))
(targ-op "##SPECIAL?"            (targ-IFJUMP-simp "SPECIALP"))
(targ-op "##PAIR?"               (targ-IFJUMP-simp "PAIRP"))
(targ-op "##SUBTYPED?"           (targ-IFJUMP-simp "SUBTYPEDP"))
(targ-op "##PROCEDURE?"          (targ-IFJUMP-simp "PROCEDUREP"))
(targ-op "##PLACEHOLDER?"        (targ-IFJUMP-simp "PLACEHOLDERP"))
(targ-op "##VECTOR?"             (targ-IFJUMP-simp "VECTORP"))
(targ-op "##SYMBOL?"             (targ-IFJUMP-simp "SYMBOLP"))
(targ-op "##RATNUM?"             (targ-IFJUMP-simp "RATNUMP"))
(targ-op "##CPXNUM?"             (targ-IFJUMP-simp "CPXNUMP"))
(targ-op "##STRING?"             (targ-IFJUMP-simp "STRINGP"))
(targ-op "##BIGNUM?"             (targ-IFJUMP-simp "BIGNUMP"))
(targ-op "##CHAR?"               (targ-IFJUMP-simp "CHARP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##FIXNUM.+"            (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXADD"))
(targ-op "##FIXNUM.*"            (targ-APPLY-fold "FIX1"  "FIXPOS" "FIXMUL"))
(targ-op "##FIXNUM.-"            (targ-APPLY-fold #f      "FIXNEG" "FIXSUB"))
(targ-op "##FIXNUM.QUOTIENT"     (targ-APPLY-fold #f      "FIXINV" "FIXQUO"))
(targ-op "##FIXNUM.REMAINDER"    (targ-APPLY-simp #f "FIXREM"))
(targ-op "##FIXNUM.MODULO"       (targ-APPLY-simp #f "FIXMOD"))
(targ-op "##FIXNUM.LOGIOR"       (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXIOR"))
(targ-op "##FIXNUM.LOGXOR"       (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXXOR"))
(targ-op "##FIXNUM.LOGAND"       (targ-APPLY-fold "FIXM1" "FIXPOS" "FIXAND"))
(targ-op "##FIXNUM.LOGNOT"       (targ-APPLY-simp #f "FIXNOT"))
(targ-op "##FIXNUM.ASH"          (targ-APPLY-simp #f "FIXASH"))
(targ-op "##FIXNUM.LSH"          (targ-APPLY-simp #f "FIXLSH"))

(targ-op "##FIXNUM.ZERO?"        (targ-IFJUMP-simp "FIXZEROP"))
(targ-op "##FIXNUM.POSITIVE?"    (targ-IFJUMP-simp "FIXPOSITIVEP"))
(targ-op "##FIXNUM.NEGATIVE?"    (targ-IFJUMP-simp "FIXNEGATIVEP"))
(targ-op "##FIXNUM.ODD?"         (targ-IFJUMP-simp "FIXODDP"))
(targ-op "##FIXNUM.EVEN?"        (targ-IFJUMP-simp "FIXEVENP"))
(targ-op "##FIXNUM.="            (targ-IFJUMP-fold "FIXEQ"))
(targ-op "##FIXNUM.<"            (targ-IFJUMP-fold "FIXLT"))
(targ-op "##FIXNUM.>"            (targ-IFJUMP-fold "FIXGT"))
(targ-op "##FIXNUM.<="           (targ-IFJUMP-fold "FIXLE"))
(targ-op "##FIXNUM.>="           (targ-IFJUMP-fold "FIXGE"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##FLONUM.->FIXNUM"     (targ-APPLY-simp #f "FLOTOFIX"))
(targ-op "##FLONUM.<-FIXNUM"     (targ-APPLY-simpflo "FLOFROMFIX"))
(targ-op "##FLONUM.+"            (targ-APPLY-foldflo "FLO0" "FLOPOS" "FLOADD"))
(targ-op "##FLONUM.*"            (targ-APPLY-foldflo "FLO1" "FLOPOS" "FLOMUL"))
(targ-op "##FLONUM.-"            (targ-APPLY-foldflo #f     "FLONEG" "FLOSUB"))
(targ-op "##FLONUM./"            (targ-APPLY-foldflo #f     "FLOINV" "FLODIV"))
(targ-op "##FLONUM.ABS"          (targ-APPLY-simpflo "FLOABS"))
(targ-op "##FLONUM.FLOOR"        (targ-APPLY-simpflo "FLOFLOOR"))
(targ-op "##FLONUM.CEILING"      (targ-APPLY-simpflo "FLOCEILING"))
(targ-op "##FLONUM.ROUND"        (targ-APPLY-simpflo "FLOROUND"))
(targ-op "##FLONUM.EXP"          (targ-APPLY-simpflo "FLOEXP"))
(targ-op "##FLONUM.LOG"          (targ-APPLY-simpflo "FLOLOG"))
(targ-op "##FLONUM.SIN"          (targ-APPLY-simpflo "FLOSIN"))
(targ-op "##FLONUM.COS"          (targ-APPLY-simpflo "FLOCOS"))
(targ-op "##FLONUM.TAN"          (targ-APPLY-simpflo "FLOTAN"))
(targ-op "##FLONUM.ASIN"         (targ-APPLY-simpflo "FLOASIN"))
(targ-op "##FLONUM.ACOS"         (targ-APPLY-simpflo "FLOACOS"))
(targ-op "##FLONUM.ATAN"         (targ-APPLY-simpflo "FLOATAN"))
(targ-op "##FLONUM.SQRT"         (targ-APPLY-simpflo "FLOSQRT"))

(targ-op "##FLONUM.ZERO?"        (targ-IFJUMP-simp "FLOZEROP"))
(targ-op "##FLONUM.NEGATIVE?"    (targ-IFJUMP-simp "FLONEGATIVEP"))
(targ-op "##FLONUM.POSITIVE?"    (targ-IFJUMP-simp "FLOPOSITIVEP"))
(targ-op "##FLONUM.="            (targ-IFJUMP-fold "FLOEQ"))
(targ-op "##FLONUM.<"            (targ-IFJUMP-fold "FLOLT"))
(targ-op "##FLONUM.>"            (targ-IFJUMP-fold "FLOGT"))
(targ-op "##FLONUM.<="           (targ-IFJUMP-fold "FLOLE"))
(targ-op "##FLONUM.>="           (targ-IFJUMP-fold "FLOGE"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##CHAR=?"              (targ-IFJUMP-fold "CHAREQP"))
(targ-op "##CHAR<?"              (targ-IFJUMP-fold "CHARLTP"))
(targ-op "##CHAR>?"              (targ-IFJUMP-fold "CHARGTP"))
(targ-op "##CHAR<=?"             (targ-IFJUMP-fold "CHARLEP"))
(targ-op "##CHAR>=?"             (targ-IFJUMP-fold "CHARGEP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##CONS"                (targ-APPLY-cons))
(targ-op "##SET-CAR!"            (targ-APPLY-simp #t "SETCAR"))
(targ-op "##SET-CDR!"            (targ-APPLY-simp #t "SETCDR"))
(targ-op "##CAR"                 (targ-IFJUMP-APPLY "CAR"))
(targ-op "##CDR"                 (targ-IFJUMP-APPLY "CDR"))
(targ-op "##CAAR"                (targ-IFJUMP-APPLY "CAAR"))
(targ-op "##CADR"                (targ-IFJUMP-APPLY "CADR"))
(targ-op "##CDAR"                (targ-IFJUMP-APPLY "CDAR"))
(targ-op "##CDDR"                (targ-IFJUMP-APPLY "CDDR"))
(targ-op "##CAAAR"               (targ-IFJUMP-APPLY "CAAAR"))
(targ-op "##CAADR"               (targ-IFJUMP-APPLY "CAADR"))
(targ-op "##CADAR"               (targ-IFJUMP-APPLY "CADAR"))
(targ-op "##CADDR"               (targ-IFJUMP-APPLY "CADDR"))
(targ-op "##CDAAR"               (targ-IFJUMP-APPLY "CDAAR"))
(targ-op "##CDADR"               (targ-IFJUMP-APPLY "CDADR"))
(targ-op "##CDDAR"               (targ-IFJUMP-APPLY "CDDAR"))
(targ-op "##CDDDR"               (targ-IFJUMP-APPLY "CDDDR"))
(targ-op "##CAAAAR"              (targ-IFJUMP-APPLY "CAAAAR"))
(targ-op "##CAAADR"              (targ-IFJUMP-APPLY "CAAADR"))
(targ-op "##CAADAR"              (targ-IFJUMP-APPLY "CAADAR"))
(targ-op "##CAADDR"              (targ-IFJUMP-APPLY "CAADDR"))
(targ-op "##CADAAR"              (targ-IFJUMP-APPLY "CADAAR"))
(targ-op "##CADADR"              (targ-IFJUMP-APPLY "CADADR"))
(targ-op "##CADDAR"              (targ-IFJUMP-APPLY "CADDAR"))
(targ-op "##CADDDR"              (targ-IFJUMP-APPLY "CADDDR"))
(targ-op "##CDAAAR"              (targ-IFJUMP-APPLY "CDAAAR"))
(targ-op "##CDAADR"              (targ-IFJUMP-APPLY "CDAADR"))
(targ-op "##CDADAR"              (targ-IFJUMP-APPLY "CDADAR"))
(targ-op "##CDADDR"              (targ-IFJUMP-APPLY "CDADDR"))
(targ-op "##CDDAAR"              (targ-IFJUMP-APPLY "CDDAAR"))
(targ-op "##CDDADR"              (targ-IFJUMP-APPLY "CDDADR"))
(targ-op "##CDDDAR"              (targ-IFJUMP-APPLY "CDDDAR"))
(targ-op "##CDDDDR"              (targ-IFJUMP-APPLY "CDDDDR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##MAKE-CELL"           (targ-APPLY-make-cell))
(targ-op "##CELL-REF"            (targ-IFJUMP-APPLY "CELLREF"))
(targ-op "##CELL-SET!"           (targ-APPLY-simp #t "CELLSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##VECTOR"              (targ-APPLY-vector 'VECTOR))
(targ-op "##VECTOR-LENGTH"       (targ-APPLY-simp #f "VECTORLENGTH"))
(targ-op "##VECTOR-REF"          (targ-IFJUMP-APPLY "VECTORREF"))
(targ-op "##VECTOR-SET!"         (targ-APPLY-simp #t "VECTORSET"))
(targ-op "##VECTOR-SHRINK!"      (targ-APPLY-simp #t "VECTORSHRINK"))

(targ-op "##STRING"              (targ-APPLY-vector 'STRING))
(targ-op "##STRING-LENGTH"       (targ-APPLY-simp #f "STRINGLENGTH"))
(targ-op "##STRING-REF"          (targ-APPLY-simp #f "STRINGREF"))
(targ-op "##STRING-SET!"         (targ-APPLY-simp #t "STRINGSET"))
(targ-op "##STRING-SHRINK!"      (targ-APPLY-simp #t "STRINGSHRINK"))

(targ-op "##VECTOR8"             (targ-APPLY-vector 'VECTOR8))
(targ-op "##VECTOR8-LENGTH"      (targ-APPLY-simp #f "VECTOR8LENGTH"))
(targ-op "##VECTOR8-REF"         (targ-APPLY-simp #f "VECTOR8REF"))
(targ-op "##VECTOR8-SET!"        (targ-APPLY-simp #t "VECTOR8SET"))
(targ-op "##VECTOR8-SHRINK!"     (targ-APPLY-simp #t "VECTOR8SHRINK"))

(targ-op "##VECTOR16"            (targ-APPLY-vector 'VECTOR16))
(targ-op "##VECTOR16-LENGTH"     (targ-APPLY-simp #f "VECTOR16LENGTH"))
(targ-op "##VECTOR16-REF"        (targ-APPLY-simp #f "VECTOR16REF"))
(targ-op "##VECTOR16-SET!"       (targ-APPLY-simp #t "VECTOR16SET"))
(targ-op "##VECTOR16-SHRINK!"    (targ-APPLY-simp #t "VECTOR16SHRINK"))

(targ-op "##CLOSURE-CODE"        (targ-APPLY-simp #f "CLOSURECODE"))
(targ-op "##CLOSURE-REF"         (targ-APPLY-simp #f "CLOSUREREF"))
(targ-op "##CLOSURE-SET!"        (targ-APPLY-simp #t "CLOSURESET"))

(targ-op "##GLOBAL-VAR-REF"      (targ-APPLY-simp #f "GLOBALVARREF"))
(targ-op "##GLOBAL-VAR-SET!"     (targ-APPLY-simp #t "GLOBALVARSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##PSTATE"              (targ-APPLY-simp #f "PSTATE"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##MAKE-PLACEHOLDER"    (targ-APPLY-make-placeholder))
(targ-op "##TOUCH"               (targ-APPLY-touch))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op
  "##C-CODE"
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (targ-use-all-res)
        (let ((n (length opnds)))
          (if (= n 0)
            (compiler-error
              "\"##c-code\" needs at least one argument")
            (let ((code (car opnds))
                  (args (map (lambda (opnd) (targ-opnd opnd sn #t))
                             (cdr opnds))))
              (if (and (obj? code)
                       (string? (obj-val code)))
                (let loop ((i (- n 1))
                           (rev-args (reverse args))
                           (lst (list "#undef " c-id-prefix "RESULT"
                                      #\newline
                                      "#define " c-id-prefix "RESULT "
                                                 c-id-prefix "temp"
                                      #\newline
                                      (obj-val code)
                                      #\newline)))
                  (if (null? rev-args)
                    (begin
                      (targ-emit (cons 'APPEND lst))
                      (if loc
                        (targ-emit (targ-loc loc sn #f '("RESULT")))))
                    (let ((arg (car rev-args)))
                      (loop (- i 1)
                            (cdr rev-args)
                            (append
                              (list "#undef " c-id-prefix "ARG" i
                                    #\newline
                                    "#define " c-id-prefix "ARG" i " " arg
                                    #\newline)
                              lst)))))
                (compiler-error
                  "Argument 1 of \"##c-code\" must be a string constant")))))))))

(targ-op
  "##C-DECL"
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (let ((n (length opnds)))
          (if (not (= n 1))
            (compiler-error
              "\"##c-decl\" expects a single argument")
            (let ((code (car opnds)))
              (if (and (obj? code)
                       (string? (obj-val code)))
                (begin
                  (queue-put! targ-module-decl (obj-val code))
                  (if loc
                    (targ-emit (targ-loc loc sn #f '("FAL")))))
                (compiler-error
                  "Argument of \"##c-decl\" must be a string constant")))))))))

;------------------------------------------------------------------------------

; Table of procedure specializations

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "NOT"        (targ-s "##NOT"))
(targ-spec "NULL?"      (targ-s "##NULL?"))
(targ-spec "EQ?"        (targ-s "##EQ?"))

(targ-spec "PAIR?"      (targ-s "##PAIR?"))
(targ-spec "PROCEDURE?" (targ-s "##PROCEDURE?"))
(targ-spec "VECTOR?"    (targ-s "##VECTOR?"))
(targ-spec "SYMBOL?"    (targ-s "##SYMBOL?"))
(targ-spec "STRING?"    (targ-s "##STRING?"))
(targ-spec "CHAR?"      (targ-s "##CHAR?"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "ZERO?"      (targ-s-a "##FIXNUM.ZERO?"     "##FLONUM.ZERO?"))
(targ-spec "POSITIVE?"  (targ-s-a "##FIXNUM.POSITIVE?" "##FLONUM.POSITIVE?"))
(targ-spec "NEGATIVE?"  (targ-s-a "##FIXNUM.NEGATIVE?" "##FLONUM.NEGATIVE?"))
(targ-spec "ODD?"       (targ-s-a "##FIXNUM.ODD?"      #f))
(targ-spec "EVEN?"      (targ-s-a "##FIXNUM.EVEN?"     #f))

(targ-spec "+"          (targ-u-a "##FIXNUM.+"         "##FLONUM.+"))
(targ-spec "*"          (targ-u-a "##FIXNUM.*"         "##FLONUM.*"))
(targ-spec "-"          (targ-u-a "##FIXNUM.-"         "##FLONUM.-"))
(targ-spec "/"          (targ-u-a #f                   "##FLONUM./"))
(targ-spec "QUOTIENT"   (targ-u-a "##FIXNUM.QUOTIENT"  #f))
(targ-spec "REMAINDER"  (targ-u-a "##FIXNUM.REMAINDER" #f))
(targ-spec "MODULO"     (targ-u-a "##FIXNUM.MODULO"    #f))

(targ-spec "="          (targ-s-a "##FIXNUM.="         "##FLONUM.="))
(targ-spec "<"          (targ-s-a "##FIXNUM.<"         "##FLONUM.<"))
(targ-spec ">"          (targ-s-a "##FIXNUM.>"         "##FLONUM.>"))
(targ-spec "<="         (targ-s-a "##FIXNUM.<="        "##FLONUM.<="))
(targ-spec ">="         (targ-s-a "##FIXNUM.>="        "##FLONUM.>="))

(targ-spec "ABS"        (targ-u-a #f                   "##FLONUM.ABS"))
(targ-spec "FLOOR"      (targ-u-a #f                   "##FLONUM.FLOOR"))
(targ-spec "CEILING"    (targ-u-a #f                   "##FLONUM.CEILING"))
(targ-spec "TRUNCATE"   (targ-u-a #f                   "##FLONUM.TRUNCATE"))
(targ-spec "ROUND"      (targ-u-a #f                   "##FLONUM.ROUND"))
(targ-spec "EXP"        (targ-u-a #f                   "##FLONUM.EXP"))
(targ-spec "LOG"        (targ-u-a #f                   "##FLONUM.LOG"))
(targ-spec "SIN"        (targ-u-a #f                   "##FLONUM.SIN"))
(targ-spec "COS"        (targ-u-a #f                   "##FLONUM.COS"))
(targ-spec "TAN"        (targ-u-a #f                   "##FLONUM.TAN"))
(targ-spec "ASIN"       (targ-u-a #f                   "##FLONUM.ASIN"))
(targ-spec "ACOS"       (targ-u-a #f                   "##FLONUM.ACOS"))
(targ-spec "ATAN"       (targ-u-a #f                   "##FLONUM.ATAN"))
(targ-spec "SQRT"       (targ-u-a #f                   "##FLONUM.SQRT"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "CHAR=?"     (targ-s "##CHAR=?"))
(targ-spec "CHAR<?"     (targ-s "##CHAR<?"))
(targ-spec "CHAR>?"     (targ-s "##CHAR>?"))
(targ-spec "CHAR<=?"    (targ-s "##CHAR<=?"))
(targ-spec "CHAR>=?"    (targ-s "##CHAR>=?"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "CONS"       (targ-s "##CONS"))
(targ-spec "SET-CAR!"   (targ-u "##SET-CAR!"))
(targ-spec "SET-CDR!"   (targ-u "##SET-CDR!"))
(targ-spec "CAR"        (targ-u "##CAR"))
(targ-spec "CDR"        (targ-u "##CDR"))
(targ-spec "CAAR"       (targ-u "##CAAR"))
(targ-spec "CADR"       (targ-u "##CADR"))
(targ-spec "CDAR"       (targ-u "##CDAR"))
(targ-spec "CDDR"       (targ-u "##CDDR"))
(targ-spec "CAAAR"      (targ-u "##CAAAR"))
(targ-spec "CAADR"      (targ-u "##CAADR"))
(targ-spec "CADAR"      (targ-u "##CADAR"))
(targ-spec "CADDR"      (targ-u "##CADDR"))
(targ-spec "CDAAR"      (targ-u "##CDAAR"))
(targ-spec "CDADR"      (targ-u "##CDADR"))
(targ-spec "CDDAR"      (targ-u "##CDDAR"))
(targ-spec "CDDDR"      (targ-u "##CDDDR"))
(targ-spec "CAAAAR"     (targ-u "##CAAAAR"))
(targ-spec "CAAADR"     (targ-u "##CAAADR"))
(targ-spec "CAADAR"     (targ-u "##CAADAR"))
(targ-spec "CAADDR"     (targ-u "##CAADDR"))
(targ-spec "CADAAR"     (targ-u "##CADAAR"))
(targ-spec "CADADR"     (targ-u "##CADADR"))
(targ-spec "CADDAR"     (targ-u "##CADDAR"))
(targ-spec "CADDDR"     (targ-u "##CADDDR"))
(targ-spec "CDAAAR"     (targ-u "##CDAAAR"))
(targ-spec "CDAADR"     (targ-u "##CDAADR"))
(targ-spec "CDADAR"     (targ-u "##CDADAR"))
(targ-spec "CDADDR"     (targ-u "##CDADDR"))
(targ-spec "CDDAAR"     (targ-u "##CDDAAR"))
(targ-spec "CDDADR"     (targ-u "##CDDADR"))
(targ-spec "CDDDAR"     (targ-u "##CDDDAR"))
(targ-spec "CDDDDR"     (targ-u "##CDDDDR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "VECTOR"        (targ-s "##VECTOR"))
(targ-spec "VECTOR-LENGTH" (targ-u "##VECTOR-LENGTH"))
(targ-spec "VECTOR-REF"    (targ-u "##VECTOR-REF"))
(targ-spec "VECTOR-SET!"   (targ-u "##VECTOR-SET!"))

(targ-spec "STRING"        (targ-s "##STRING"))
(targ-spec "STRING-LENGTH" (targ-u "##STRING-LENGTH"))
(targ-spec "STRING-REF"    (targ-u "##STRING-REF"))
(targ-spec "STRING-SET!"   (targ-u "##STRING-SET!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "TOUCH"         (targ-s "##TOUCH"))

(targ-spec "##EQV?"        (targ-s-**eqv?))

(targ-spec "CALL-WITH-CURRENT-CONTINUATION"
           (targ-s "##CALL-WITH-CURRENT-CONTINUATION"))

;------------------------------------------------------------------------------
